def environment_variables():
    return {
        "GIT_AUTHOR_NAME": "AI Migrator",
        "GIT_AUTHOR_EMAIL": "dx-ai-tools--ai-migrator@block.xyz",
        "GIT_COMMITTER_NAME": "AI Migrator",
        "GIT_COMMITTER_EMAIL": "dx-ai-tools--ai-migrator@block.xyz",
    }



import tempfile

from ai_migrate.manifest import Manifest
from .manifest import Directory, FileGroup
from pathlib import Path


def test_valid_manifest():
    json = """
{
  "eval_target_repo_ref": "",
  "files": [
    {
      "filename": "service/src/test/kotlin/com/squareup/cash/enforcementactions/tasks/PreviouslyDeactivatedFilterTest.kt",
      "result": "?"
    },
    {
      "filename": "service/src/test/kotlin/com/squareup/cash/enforcementactions/tasks/PreviouslyNotifiedFilterTest.kt",
      "result": "?"
    }
  ],
  "system_prompt": "{project_dir}/system_prompt.md",
  "verify_cmd": "{py} {project_dir}/verify.py",
  "pre_verify_cmd": "{py} {project_dir}/verify.py --pre",
  "time": "2025-02-10T11:26:33.969758"
}
    """
    Manifest.model_validate_json(json)


def test_valid_manifest_groups():
    json = """
{
  "eval_target_repo_ref": "",
  "files": [
    {
        "files": [
          "service/src/test/kotlin/com/squareup/cash/enforcementactions/tasks/Test1.kt",
          "service/src/test/kotlin/com/squareup/cash/enforcementactions/tasks/PreviouslyDeactivatedFilterTest.kt"
        ],
        "result": "?"
    },
    {
      "filename": "service/src/test/kotlin/com/squareup/cash/enforcementactions/tasks/PreviouslyNotifiedFilterTest.kt",
      "result": "?"
    },
    {
      "dir": "service/src/test/kotlin/com/squareup/cash/enforcementactions/models",
      "result": "?"
    },
    {
      "dir": "service/src/test/kotlin/com/squareup/cash/enforcementactions/utils",
      "glob": "**/*.kt",
      "result": "?"
    }
  ],
  "system_prompt": "{project_dir}/system_prompt.md",
  "verify_cmd": "{py} {project_dir}/verify.py",
  "pre_verify_cmd": "{py} {project_dir}/verify.py --pre",
  "time": "2025-02-10T11:26:33.969758"
}
    """
    manifest = Manifest.model_validate_json(json)
    assert len(manifest.files) == 4

    # Check the structure of the group name rather than the exact hash
    group_name = manifest.files[0].group_name()
    assert group_name.startswith(
        "service__src__test__kotlin__com__squareup__cash__enforcementactions__tasks__Test1.kt-"
    )
    assert len(group_name.split("-")[1]) == 8  # Check that the hash is 8 characters

    assert (
        manifest.files[1].group_name()
        == "service__src__test__kotlin__com__squareup__cash__enforcementactions__tasks__PreviouslyNotifiedFilterTest.kt"
    )

    # Check the Directory group_name format
    dir_group_name = manifest.files[2].group_name()
    assert dir_group_name.startswith(
        "service__src__test__kotlin__com__squareup__cash__enforcementactions__models-"
    )
    assert len(dir_group_name.split("-")[1]) == 8  # Check that the hash is 8 characters

    # Check the Directory group_name with custom glob
    dir_glob_group_name = manifest.files[3].group_name()
    assert dir_glob_group_name.startswith(
        "service__src__test__kotlin__com__squareup__cash__enforcementactions__utils-"
    )
    assert (
        len(dir_glob_group_name.split("-")[1]) == 8
    )  # Check that the hash is 8 characters
    # The hash should be different because the glob pattern is different
    assert dir_glob_group_name.split("-")[1] != dir_group_name.split("-")[1]


def test_normalize_files():
    """Test that Directory.to_file_group correctly converts Directory objects to FileGroup objects."""

    # Create a temporary directory structure for testing
    with tempfile.TemporaryDirectory() as temp_dir:
        # Create a directory structure
        test_dir = Path(temp_dir) / "test_dir"
        test_dir.mkdir()

        # Create some test files with different extensions
        (test_dir / "file1.txt").write_text("test content 1")
        (test_dir / "file2.txt").write_text("test content 2")
        (test_dir / "file3.py").write_text("test content 3")

        # Create a subdirectory with files
        sub_dir = test_dir / "subdir"
        sub_dir.mkdir()
        (sub_dir / "file4.txt").write_text("test content 4")
        (sub_dir / "file5.py").write_text("test content 5")

        # Create Directory objects using different glob patterns
        dir1 = Directory(dir=str(test_dir), result="?")  # Default glob (all files)
        dir2 = Directory(
            dir=str(test_dir), glob="**/*.py", result="py_only"
        )  # Only Python files

        # Convert Directory objects to FileGroup objects
        file_group1 = dir1.to_file_group()
        file_group2 = dir2.to_file_group()

        # Check that the Directory objects were converted to FileGroup objects
        assert isinstance(file_group1, FileGroup)
        assert isinstance(file_group2, FileGroup)

        # Check that the first FileGroup contains all files (default glob)
        assert len(file_group1.files) == 5

        # Check that the second FileGroup contains only Python files
        assert len(file_group2.files) == 2
        py_files = set(file_group2.files)
        expected_py_files = {str(test_dir / "file3.py"), str(sub_dir / "file5.py")}
        assert py_files == expected_py_files

        # Check that the result was preserved
        assert file_group1.result == "?"
        assert file_group2.result == "py_only"


def test_directory_to_file_group():
    """Test that Directory.to_file_group correctly converts a Directory to a FileGroup."""
    from .manifest import Directory, flatten, sha256
    import tempfile
    from pathlib import Path

    # Create a temporary directory structure for testing
    with tempfile.TemporaryDirectory() as temp_dir:
        # Create a directory structure
        test_dir = Path(temp_dir) / "test_dir"
        test_dir.mkdir()

        # Create some test files with different extensions
        (test_dir / "file1.txt").write_text("test content 1")
        (test_dir / "file2.txt").write_text("test content 2")
        (test_dir / "file3.py").write_text("test content 3")

        # Create a subdirectory with files
        sub_dir = test_dir / "subdir"
        sub_dir.mkdir()
        (sub_dir / "file4.txt").write_text("test content 4")
        (sub_dir / "file5.py").write_text("test content 5")

        # Test the group_name method
        directory = Directory(dir=str(test_dir), result="test_result")
        expected_hash = sha256(f"{str(test_dir)}:**/*".encode()).hexdigest()[:8]
        expected_group_name = f"{flatten(str(test_dir))}-{expected_hash}"
        assert directory.group_name() == expected_group_name

        # Test with a different glob pattern
        directory_py = Directory(dir=str(test_dir), glob="**/*.py", result="py_only")
        expected_hash_py = sha256(f"{str(test_dir)}:**/*.py".encode()).hexdigest()[:8]
        expected_group_name_py = f"{flatten(str(test_dir))}-{expected_hash_py}"
        assert directory_py.group_name() == expected_group_name_py

        # Test with default glob pattern (all files)
        directory = Directory(dir=str(test_dir), result="test_result")
        file_group = directory.to_file_group()

        # Check that the FileGroup contains all files
        assert len(file_group.files) == 5

        # Check that the paths are correct
        file_paths = set(file_group.files)
        expected_paths = {
            str(test_dir / "file1.txt"),
            str(test_dir / "file2.txt"),
            str(test_dir / "file3.py"),
            str(sub_dir / "file4.txt"),
            str(sub_dir / "file5.py"),
        }
        assert file_paths == expected_paths

        # Check that the result was preserved
        assert file_group.result == "test_result"

        # Test with a specific glob pattern (only Python files)
        directory_py = Directory(dir=str(test_dir), glob="**/*.py", result="py_only")
        file_group_py = directory_py.to_file_group()

        # Check that the FileGroup contains only Python files
        assert len(file_group_py.files) == 2

        # Check that the paths are correct
        py_file_paths = set(file_group_py.files)
        expected_py_paths = {str(test_dir / "file3.py"), str(sub_dir / "file5.py")}
        assert py_file_paths == expected_py_paths

        # Check that the result was preserved
        assert file_group_py.result == "py_only"

        # Test with a more specific glob pattern (only txt files in the root directory)
        directory_txt_root = Directory(
            dir=str(test_dir), glob="*.txt", result="txt_root"
        )
        file_group_txt_root = directory_txt_root.to_file_group()

        # Check that the FileGroup contains only txt files in the root directory
        assert len(file_group_txt_root.files) == 2

        # Check that the paths are correct
        txt_root_file_paths = set(file_group_txt_root.files)
        expected_txt_root_paths = {
            str(test_dir / "file1.txt"),
            str(test_dir / "file2.txt"),
        }
        assert txt_root_file_paths == expected_txt_root_paths


def test_directory_with_multiple_extensions():
    """Test that Directory.to_file_group correctly handles glob patterns with multiple extensions."""
    from .manifest import Directory
    import tempfile
    from pathlib import Path

    # Create a temporary directory structure for testing
    with tempfile.TemporaryDirectory() as temp_dir:
        # Create a directory structure
        test_dir = Path(temp_dir) / "test_dir"
        test_dir.mkdir()

        # Create files with different extensions
        (test_dir / "file1.kt").write_text("kotlin file")
        (test_dir / "file2.js").write_text("javascript file")
        (test_dir / "file3.py").write_text("python file")
        (test_dir / "file4.txt").write_text("text file")

        # Create a subdirectory with files
        sub_dir = test_dir / "subdir"
        sub_dir.mkdir()
        (sub_dir / "file5.kt").write_text("another kotlin file")
        (sub_dir / "file6.js").write_text("another javascript file")
        (sub_dir / "file7.py").write_text("another python file")

        # Test with a glob pattern that matches multiple extensions
        directory = Directory(
            dir=str(test_dir), glob="**/*.{kt,js}", result="kt_js_files"
        )
        file_group = directory.to_file_group()

        # Print debug information
        print(f"Directory: {test_dir}")
        print(f"Files found: {file_group.files}")

        # Check that the FileGroup contains only .kt and .js files
        assert len(file_group.files) == 4

        # Check that the paths are correct
        file_paths = set(file_group.files)
        expected_paths = {
            str(test_dir / "file1.kt"),
            str(test_dir / "file2.js"),
            str(sub_dir / "file5.kt"),
            str(sub_dir / "file6.js"),
        }
        assert file_paths == expected_paths

        # Check that the result was preserved
        assert file_group.result == "kt_js_files"


def test_manifest_with_target_dir():
    """Test that a manifest with a target_dir is correctly parsed."""
    from .manifest import Manifest

    json = """
{
  "eval_target_repo_ref": "",
  "migrate_repo_ref": "",
  "files": [
    {
      "filename": "src/main/java/com/example/App.java",
      "result": "?"
    }
  ],
  "target_dir": "/path/to/target/repo",
  "system_prompt": "{project_dir}/system_prompt.md",
  "verify_cmd": "{py} {project_dir}/verify.py",
  "pre_verify_cmd": "{py} {project_dir}/verify.py --pre",
  "time": "2025-02-10T11:26:33.969758"
}
    """
    manifest = Manifest.model_validate_json(json)
    assert manifest.target_dir == "/path/to/target/repo"
    assert len(manifest.files) == 1
    assert manifest.files[0].filename == "src/main/java/com/example/App.java"



import asyncio
import re
import subprocess
import traceback
from datetime import datetime
from pathlib import Path
import sys
from typing import Iterable

from ai_migrate.git import get_branches
from .manifest import (
    FileEntry,
    Manifest,
    FileGroup,
    Directory,
)
from .migrate import run as run_migration, FailedPreVerification
from .progress import StatusManager


def get_git_sha(directory: str | Path) -> str:
    """Get the git SHA for HEAD of the repo in the current directory"""
    res = subprocess.run(
        ["git", "rev-parse", "HEAD"],
        cwd=str(directory),
        check=True,
        capture_output=True,
    )
    return res.stdout.decode().strip()


def normalize_file_group(
    file_dir_group: FileEntry | FileGroup | Directory,
) -> FileGroup:
    if isinstance(file_dir_group, FileGroup):
        return file_dir_group
    if isinstance(file_dir_group, Directory):
        return file_dir_group.to_file_group()
    return FileGroup(
        files=[file_dir_group.filename],
        result=file_dir_group.result,
    )


def manifest_from_git() -> Manifest:
    files = []
    for _sha, branch, status, _msg in get_branches():
        file_name = branch.split("/", maxsplit=1)[1].replace("__", "/")
        files.append(FileEntry(filename=file_name, result=status))
    return Manifest(files=files)


def merge_manifests(manifest1: Manifest, manifest2: Manifest):
    entries = {f.group_name(): f for f in manifest1.files}
    for f in manifest2.files:
        if f.result == "pass" and f.group_name() in entries:
            entries[f.group_name()].result = "pass"
    return manifest1.model_copy(update={"files": list(entries.values())})


async def run(
    project_dir: str,
    logs_dir: str | Path,
    manifest_file: str | None,
    files: list[str] | None,
    only_failed: bool,
    max_workers: int = 8,
    local_worktrees=False,
    resume: bool = True,
    llm_fakes=None,
    dont_create_evals: bool = False,
) -> list[FileGroup]:
    """Run an AI migration project."""
    if manifest_file:
        with open(manifest_file) as f:
            manifest = Manifest.model_validate_json(f.read())
    else:
        manifest = Manifest(files=[])

    if resume:
        git_manifest = manifest_from_git()
        manifest = merge_manifests(manifest, git_manifest)

    target_sha = None

    results: list[FileGroup] = []
    if files:
        file_list = [FileGroup(files=[f], result="?") for f in files]
    else:
        file_list = [normalize_file_group(f) for f in manifest.files]

    if not file_list:
        print(
            "No files to migrate. Provide a non-empty manifest or specify files on the command line. "
            "Use the create_manifest command to set up an initial file manifest."
        )
        return []

    if only_failed:
        file_list = [f for f in file_list if f.result != "pass"]

    logs_dir = Path(logs_dir)

    status_manager = StatusManager()

    async def process_one_fileset(index, files: FileGroup, task_name: str):
        nonlocal target_sha
        if target_sha is None:
            target_sha = get_git_sha(Path(files.files[0]).parent)

        await status_manager.update_message(task_name, "Running...")

        log_file = (logs_dir / task_name).with_suffix(".log")
        log_file.parent.mkdir(parents=True, exist_ok=True)

        log_buffer = open(log_file, "w")
        try:
            await run_migration(
                files.files,
                manifest.system_prompt.format(project_dir=project_dir),
                Path(project_dir) / "examples",
                verify_cmd=manifest.verify_cmd.format(
                    project_dir=project_dir, py=sys.executable
                ),
                pre_verify_cmd=manifest.pre_verify_cmd.format(
                    project_dir=project_dir, py=sys.executable
                ),
                log_stream=log_buffer,
                local_worktrees=local_worktrees,
                llm_fakes=llm_fakes,
                dont_create_evals=dont_create_evals,
                target_dir=manifest.target_dir,
                target_basename=files.base_name,
            )
            new_result = "pass"
            await status_manager.mark_passed(task_name)
        except FailedPreVerification:
            await status_manager.mark_failed(task_name)
            new_result = "fail-pre-verify"
        except Exception:
            await status_manager.mark_failed(task_name)
            traceback.print_exc(file=log_buffer)
            new_result = "fail"
        finally:
            await status_manager.update_message(task_name, "")
            log_buffer.close()

        results.append(FileGroup(files=files.files, result=new_result))

    sem = asyncio.Semaphore(max_workers)

    async def process_one_with_sem(index, files: FileGroup, task_name: str):
        async with sem:
            try:
                await process_one_fileset(index, files, task_name)
            except Exception:
                print("Unexpected error in task", task_name)
                traceback.print_exc()

    async with asyncio.TaskGroup() as tg:
        for i, file_set in enumerate(file_list):
            task_name = Path(file_set.files[0]).name
            if len(file_set.files) > 1:
                task_name = task_name + f" (+{len(file_set.files) - 1})"

            await status_manager.add_status(task_name)
            await status_manager.update_message(task_name, "Waiting...")
            tg.create_task(process_one_with_sem(i, file_set, task_name))

    print("Project run complete.")
    print("Failing files:")
    for file in results:
        if "fail" in file.result:
            for fn in file.files:
                print(fn)
    print("Passing files:")
    for file in results:
        if file.result == "pass":
            for fn in file.files:
                print(fn)

    ts = datetime.now().strftime("%Y%m%d-%H%M%S")
    results_file = f"manifest-{ts}.json"
    with open(results_file, "w") as f:
        result_manifest = manifest.model_copy(
            update={
                "files": results,
                "eval_target_repo_ref": target_sha,
                "time": datetime.now(),
            }
        )
        f.write(
            result_manifest.model_dump_json(
                indent=2, exclude_defaults=True, exclude_none=True
            )
        )

    print(f"Results saved to {results_file}")
    return results


def verify(project_dir: str, files: Iterable[str], manifest_file: str | None):
    if manifest_file:
        manifest = Manifest.model_validate_json(Path(manifest_file).read_text())
    else:
        manifest = Manifest()
    verify_cmd = manifest.verify_cmd.format(project_dir=project_dir, py=sys.executable)
    subprocess.run([*verify_cmd.split(), *files], check=True)


def pre_verify(project_dir: str, file: str, manifest_file: str | None):
    if manifest_file:
        manifest = Manifest.model_validate_json(Path(manifest_file).read_text())
    else:
        manifest = Manifest()
    verify_cmd = manifest.pre_verify_cmd.format(
        project_dir=project_dir, py=sys.executable
    )
    subprocess.run([*verify_cmd.split(), file], check=True)


def status(manifest_file: str):
    manifest = Manifest.model_validate_json(Path(manifest_file).read_text())
    branches = get_branches(manifest)
    passing = []
    failing = []
    failed_to_start = []

    for ref, branch, status, msg in branches:
        if status == "pass":
            passing.append((ref, branch, msg))
        elif "fail" in status:
            failing.append((ref, branch, msg))
        else:
            failed_to_start.append((ref, branch, msg))

    def branch_to_filename(branch: str):
        filename = branch.removeprefix("ai-migrator/")
        parts = filename.split("__")
        if len(parts) > 3:
            parts = parts[:1] + ["..."] + parts[-2:]
        return "/".join(parts)

    def extract_attempts(msg: str) -> int | None:
        match = re.search(r"Migration attempt (\d+) status=", msg)
        if match:
            return match.group(1)

    def print_list(lst, status):
        if lst:
            print(f"{len(lst)} files {status}:")
            print(
                "\n".join(
                    f"{branch_to_filename(b)} ({extract_attempts(msg) or 'no'} attempts): {ref}"
                    for ref, b, msg in lst
                )
            )

    if not failing and not failed_to_start:
        print("All files passing! Ready to merge.")
    else:
        print_list(passing, "passing")
        print_list(failing, "failing")
        print_list(failed_to_start, "failed to start")


def checkout_failed(file, examples_dir="examples") -> None:
    branches = get_branches()
    matching = []
    for _sha, branch, _status, _msg in branches:
        if file in branch:
            matching.append(branch)
    if not matching:
        raise ValueError(f"No failed branches found for {file}")
    if len(matching) > 1:
        raise ValueError(f"Multiple failed branches found for {file}: {matching}")
    (failed_branch,) = matching
    subprocess.run(
        ["git", "checkout", "-b", f"repair-{failed_branch}", failed_branch], check=True
    )
    print("Checked out", failed_branch)
    print("Use `git log` to see what the migrator tried.")



import json
import os
import re
import subprocess

REPOSITORY_TREE = None


def contains_definition(line: str, symbol: str) -> bool:
    line = line.strip()
    if re.match(
        rf"^(public|private|protected|static|final|abstract|synchronized|data|\s)*"
        rf"(class|interface|enum|void|[a-zA-Z_][a-zA-Z0-9_<,>\[\]]*)\s+{re.escape(symbol)}(\s*<[^>]+>)?\b",
        line,
    ):
        return True
    return bool(
        re.match(
            rf"^(public|private|protected|static|final|\s)*"
            rf"[a-zA-Z_][a-zA-Z0-9_<,>\[\]]*\s+{re.escape(symbol)}\s*\(.*\)",
            line,
        )
    )


def find_symbol_definition(file_path: str, symbol_name: str) -> str | None:
    in_comment = False
    in_definition = False
    current_comment = []
    current_definition = []
    # Hacky way to handle nested symbols:
    symbol_bits = symbol_name.split(".")

    with open(file_path, "r", encoding="utf-8") as file:
        for line in file:
            if not line.strip():
                continue

            line = line.rstrip()
            if in_comment:
                if (p := line.find("*/")) != -1:
                    current_comment.append(line[: p + 2])
                    in_comment = False
                    continue
                else:
                    current_comment.append(line)
                    continue
            if "//" in line:
                line, comment = line.split("//", 1)
                current_comment.append(comment)

            if (p := line.find("/*")) != -1:
                endp = line.find("*/", p)
                current_comment = [line[p:] if endp == -1 else line[p : endp + 2]]
                line = line[:p]
                in_comment = endp == -1

            if symbol_bits and contains_definition(line, symbol_bits[0]):
                symbol_bits.pop(0)
                in_definition = len(symbol_bits) == 0

            if not in_definition:
                current_comment = []
                continue

            if "{" in line or ";" in line:
                line = line.split("{")[0].split(";")[0].strip()
                all_lines = [*current_comment, *current_definition, line]
                return "\n".join(all_lines)

            current_definition.append(line)
    return None


def get_symbol_definition(symbol: str, package: str) -> str | None:
    global REPOSITORY_TREE
    if REPOSITORY_TREE is None:
        REPOSITORY_TREE = scan_repository()
    root = REPOSITORY_TREE
    for bit in package.split("."):
        if bit not in root:
            break
        root = root[bit]
    for file_path in root.get("__files__", []):
        definition = find_symbol_definition(file_path, symbol)
        if definition:
            return definition
    return None


def scan_repository(root_dir: str | None = None) -> dict[str, str]:
    if root_dir is None:
        root_dir = subprocess.run(
            ["git", "rev-parse", "--show-toplevel"],
            capture_output=True,
            text=True,
        ).stdout.strip()
    cache_path = os.path.join(root_dir, "symbol_tree.json")
    if os.path.exists(cache_path):
        with open(cache_path, "r") as f:
            return json.load(f)
    print("Building symbol tree for repository at", root_dir)
    symbol_tree = {}
    for dirpath, _, filenames in os.walk(root_dir):
        for filename in filenames:
            if filename.endswith((".kt", ".java")):
                file_path = os.path.join(dirpath, filename)
                for line in open(file_path, "r"):
                    if line.startswith("package "):
                        package = line.split("package ", 1)[1].strip()
                        break
                else:
                    continue
                root = symbol_tree
                for bit in package.split("."):
                    root = root.setdefault(bit, {})
                root.setdefault("__files__", []).append(file_path)
    with open(cache_path, "w") as f:
        json.dump(symbol_tree, f, indent=2)
    return symbol_tree


if __name__ == "__main__":
    scan_repository()



import glob
import json
import re
from pathlib import Path
import subprocess


def get_git_file_content(file: str, ref: str) -> str | None:
    result = subprocess.run(
        ["git", "show", f"{ref}:{file}"],
        capture_output=True,
    )
    if result.returncode != 0:
        stderr = result.stderr.decode()
        if re.match(
            r"fatal: path '.*' exists on disk, but not in '[a-f0-9+]'",
            stderr,
        ) or re.match(
            r"fatal: path '.*' does not exist in '[a-f0-9+]'",
            stderr,
        ):
            return None
    return result.stdout.decode()


def path_to_name(path: Path | str) -> str:
    return str(path).replace("/", "__")


def setup_from_pr(pr_num, examples_dir, pattern=None) -> None:
    default_branch = (
        subprocess.run(
            [
                "git",
                "rev-parse",
                "--abbrev-ref",
                "origin/HEAD",
            ],
            capture_output=True,
            check=True,
        )
        .stdout.decode()
        .strip()
    )

    output = subprocess.run(
        [
            "gh",
            "pr",
            "view",
            str(pr_num),
            "--json",
            "mergeCommit,headRefOid",
        ],
        capture_output=True,
        check=True,
    ).stdout.decode()
    out = json.loads(output)
    merge_commit = out.get("mergeCommit")
    if merge_commit:
        ref = merge_commit["oid"]
    else:
        ref = out["headRefOid"]

    subprocess.run(["git", "fetch"], check=True)

    prev_ref = (
        subprocess.run(
            ["git", "merge-base", default_branch, f"{ref}~"],
            capture_output=True,
            check=True,
        )
        .stdout.decode()
        .strip()
    )

    return setup(ref, prev_ref, examples_dir, pattern)


def setup(ref, ref_prev=None, examples_dir="examples", pattern=None) -> None:
    ref = str(ref)
    if ref_prev is None:
        ref_prev = f"{ref}~"
    files = (
        subprocess.run(
            [
                "git",
                "diff-tree",
                "--no-commit-id",
                "--name-only",
                "-r",
                f"{ref}..{ref_prev}",
            ],
            capture_output=True,
            check=True,
        )
        .stdout.decode()
        .splitlines()
    )
    if pattern:
        files = [file for file in files if glob.fnmatch.fnmatch(file, pattern)]

    out_dir = Path(examples_dir)
    i = 0
    taken = set(o.name for o in out_dir.iterdir())
    while (f"group{i}.new" in taken) or (f"group{i}.old" in taken):
        i += 1

    group_name = f"group{i}"

    for suffix in ("new", "old"):
        d = out_dir / f"{group_name}.{suffix}"
        d.mkdir()

    for file in files:
        for get_ref, ext in (
            (ref, "new"),
            (ref_prev, "old"),
        ):
            content = get_git_file_content(file, get_ref)
            if content:
                out_file = out_dir / f"{group_name}.{ext}" / path_to_name(file)
                out_file.write_text(content)



from datetime import datetime
from hashlib import sha256
from pathlib import Path

from pydantic import BaseModel, Field

SYSTEM_PROMPT_FILE = "system_prompt.md"
VERIFY_SCRIPT_FILE = "verify.py"


def flatten(filename):
    return filename.replace("/", "__")


class FileEntry(BaseModel):
    filename: str
    result: str = "?"

    def group_name(self) -> str:
        return flatten(self.filename)


class FileGroup(BaseModel):
    files: list[str]
    result: str = "?"
    base_name: str = ""  # Computed from directory

    def group_name(self) -> str:
        if len(self.files) == 1:
            return FileEntry(filename=self.files[0]).group_name()
        hsh = sha256(",".join(sorted(self.files)).encode()).hexdigest()[:8]
        return f"{flatten(self.files[0])}-{hsh}"


class Directory(BaseModel):
    dir: str
    glob: str = "**/*"  # Default to all files
    result: str = "?"

    def group_name(self) -> str:
        hsh = sha256(f"{self.dir}:{self.glob}".encode()).hexdigest()[:8]
        return f"{flatten(self.dir)}-{hsh}"

    def to_file_group(self) -> FileGroup:
        """Convert this Directory to a FileGroup by recursively finding all files.

        If a glob pattern is specified, only files matching the pattern will be included.
        Supports brace expansion patterns like "**/*.{kt,js}" for multiple extensions.
        """
        files = []
        dir_path = Path(self.dir)

        if dir_path.exists() and dir_path.is_dir():
            if "{" in self.glob and "}" in self.glob:
                prefix, suffix = self.glob.split("{", 1)
                extensions = suffix.split("}", 1)[0].split(",")
                remainder = suffix.split("}", 1)[1] if "}" in suffix else ""
                patterns = [f"{prefix}{ext}{remainder}" for ext in extensions]

                for pattern in patterns:
                    for item in dir_path.glob(pattern):
                        if item.is_file():
                            files.append(str(item))
            else:
                for item in dir_path.glob(self.glob):
                    if item.is_file():
                        files.append(str(item))

        base_name = self.dir.split("/")[-1]
        return FileGroup(files=files, result=self.result, base_name=base_name)


class Manifest(BaseModel):
    eval_target_repo_ref: str = ""
    eval_target_repo_remote: str = ""
    # noinspection PyDataclass
    files: list[FileGroup | FileEntry | Directory] = []
    target_dir: str = ""
    system_prompt: str = f"{{project_dir}}/{SYSTEM_PROMPT_FILE}"
    verify_cmd: str = f"{{py}} {{project_dir}}/{VERIFY_SCRIPT_FILE}"
    pre_verify_cmd: str = f"{{py}} {{project_dir}}/{VERIFY_SCRIPT_FILE} --pre"
    time: datetime = Field(default_factory=datetime.now)



from importlib.metadata import entry_points

from .openai import OpenAIClient
from .base import BaseLLMClient

try:
    DefaultClient = entry_points(group="ai_migrate")["default_llm_provider"].load()
except KeyError:
    DefaultClient = OpenAIClient

__all__ = ["BaseLLMClient", "DefaultClient", "OpenAIClient"]



"""
Interactive CLI for AI Migrate using Click, Rich, and Prompt_toolkit.

Environment Variables:
- AI_MIGRATE_PROJECT_DIR: Path to the project directory
- AI_MIGRATE_PROJECT_NAME: Project name (required when not in interactive mode)
- AI_MIGRATE_PROJECT_PATH: Project path (required when not in interactive mode)
- AI_MIGRATE_OPTION: Option to select in non-interactive mode for radiolist choices
- AI_MIGRATE_YES_NO: Option to select in non-interactive mode for yes/no choices ('yes' or 'no')
"""

import os
import re
import sys
import json
import asyncio
import subprocess
import tempfile
from datetime import datetime
from pathlib import Path
from typing import Callable, Any

import click
from rich.console import Console
from rich.panel import Panel
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.table import Table
from rich.markdown import Markdown
from prompt_toolkit import prompt
from prompt_toolkit.completion import PathCompleter

from ai_migrate.git import get_branches, get_worktrees
from ai_migrate.projects_root import get_project_dir
from .pr_utils import (
    setup_project_from_pr,
    get_pr_details,
    extract_example_patterns,
    save_examples,
    generate_system_prompt,
)
from .manifest import SYSTEM_PROMPT_FILE, Manifest
from .migrate import SYSTEM_MESSAGE
from .eval_generator import generate_eval_from_pr
from ai_migrate.llm_providers import DefaultClient
from .examples import setup as setup_examples, setup_from_pr

IS_INTERACTIVE = sys.stdin.isatty()

console = Console()


def print_header():
    """Print a stylized header for the CLI."""
    console.print(
        Panel.fit(
            "[bold blue]AI Migrate[/bold blue] - Code Migration Tool",
            border_style="blue",
        )
    )
    console.print()


def console_radiolist(title: str, text: str, values: list) -> str:
    """Console-based alternative to radiolist_dialog."""
    console.print(f"[bold blue]{title}[/bold blue]")
    console.print(text)
    console.print()

    for i, (value, description) in enumerate(values, 1):
        console.print(f"[bold cyan]{i}.[/bold cyan] {value} - {description}")

    console.print()

    if not IS_INTERACTIVE:
        option = os.environ.get("AI_MIGRATE_OPTION")
        if option:
            for value, _ in values:
                if value == option:
                    console.print(
                        f"[yellow]Not in interactive terminal. Using provided option: {option}[/yellow]"
                    )
                    return value

            console.print(
                f"[red]Error: Provided option '{option}' is not valid. Available options: {[v[0] for v in values]}[/red]"
            )
            return None
        else:
            console.print(
                "[red]Error: Not in interactive terminal and no option provided via AI_MIGRATE_OPTION environment variable.[/red]"
            )
            return None

    while True:
        choice = prompt(f"Enter your choice (1-{len(values)}, or 'q' to cancel): ")
        if choice.lower() == "q":
            return None

        try:
            index = int(choice) - 1
            if 0 <= index < len(values):
                return values[index][0]
            else:
                console.print("[yellow]Invalid choice. Please try again.[/yellow]")
        except ValueError:
            console.print("[yellow]Please enter a number or 'q' to cancel.[/yellow]")


def console_yes_no(title: str, text: str) -> bool:
    """Console-based alternative to yes_no_dialog."""
    console.print(f"[bold blue]{title}[/bold blue]")
    console.print(text)
    console.print()

    if not IS_INTERACTIVE:
        option = os.environ.get("AI_MIGRATE_YES_NO")
        if option:
            if option.lower() in ("y", "yes", "true"):
                console.print(
                    "[yellow]Not in interactive terminal. Using provided option: yes[/yellow]"
                )
                return True
            elif option.lower() in ("n", "no", "false"):
                console.print(
                    "[yellow]Not in interactive terminal. Using provided option: no[/yellow]"
                )
                return False
            else:
                console.print(
                    f"[red]Error: Provided option '{option}' is not valid. Use 'yes' or 'no'.[/red]"
                )
                return False
        else:
            console.print(
                "[red]Error: Not in interactive terminal and no option provided via AI_MIGRATE_YES_NO environment variable.[/red]"
            )
            return False

    while True:
        choice = prompt("Enter your choice (y/n): ").lower()
        if choice in ("y", "yes"):
            return True
        elif choice in ("n", "no"):
            return False
        else:
            console.print("[yellow]Please enter 'y' or 'n'.[/yellow]")


def run_with_progress(description: str, func: Callable, *args, **kwargs) -> Any:
    """Run a function with a progress indicator."""
    with Progress(
        SpinnerColumn(),
        TextColumn("[bold blue]{task.description}"),
        console=console,
    ) as progress:
        task = progress.add_task(f"[bold blue]{description}", total=None)

        try:
            result = func(*args, **kwargs)
            progress.update(task, completed=True)
            return result, None
        except Exception as e:
            progress.update(task, completed=True)
            return None, e


def run_async_with_progress(description: str, func: Callable, *args, **kwargs) -> Any:
    """Run an async function with a progress indicator."""

    async def wrapper():
        return await func(*args, **kwargs)

    with Progress(
        SpinnerColumn(),
        TextColumn("[bold blue]{task.description}"),
        console=console,
    ) as progress:
        task = progress.add_task(f"[bold blue]{description}", total=None)

        try:
            result = asyncio.run(wrapper())
            progress.update(task, completed=True)
            return result, None
        except Exception as e:
            progress.update(task, completed=True)
            return None, e


def show_success_message(message: str):
    """Show a success message."""
    console.print(f"[green]✓[/green] {message}")


def show_error_message(message: str, error=None):
    """Show an error message."""

    safe_message = message.replace("[", "\\[").replace("]", "\\]")
    console.print(f"[red]✗[/red] {safe_message}")
    if error:
        safe_error = str(error).replace("[", "\\[").replace("]", "\\]")
        console.print(f"[red]Error details:[/red] {safe_error}")


def show_warning_message(message: str):
    """Show a warning message."""
    console.print(f"[yellow]Warning:[/yellow] {message}")


def manage_examples(project_dir: Path):
    """Manage example files."""
    examples_dir = project_dir / "examples"
    if not examples_dir.exists():
        show_error_message(f"Examples directory {examples_dir} does not exist.")
        return

    action = console_radiolist(
        title="Examples Management",
        text="What would you like to do?",
        values=[
            ("list", "List existing examples"),
            ("add", "Add a new example"),
            ("from-pr", "Generate examples from a PR"),
            ("setup", "Setup examples from git history"),
        ],
    )

    if not action:
        return

    if action == "list":
        files = list(examples_dir.glob("*"))

        if not files:
            console.print("No examples found.")
            return

        table = Table(title="Example Files")
        table.add_column("Filename", style="cyan")
        table.add_column("Type", style="green")
        table.add_column("Size", justify="right", style="blue")

        for file in files:
            file_type = (
                "Old"
                if ".old." in file.name
                else "New"
                if ".new." in file.name
                else "Unknown"
            )
            table.add_row(file.name, file_type, f"{file.stat().st_size} bytes")

        console.print(table)

    elif action == "add":
        file_path = prompt(
            "Enter path to the file to add as example: ", completer=PathCompleter()
        )

        file_path = Path(file_path).absolute()
        if not file_path.exists():
            show_error_message(f"File {file_path} does not exist.")
            return

        example_type = console_radiolist(
            title="Example Type",
            text="Is this an old (before) or new (after) example?",
            values=[
                ("old", "Old (before migration)"),
                ("new", "New (after migration)"),
            ],
        )

        if not example_type:
            return

        base_name = prompt("Enter base name for the example: ", default=file_path.stem)

        target_path = examples_dir / f"{base_name}.{example_type}{file_path.suffix}"
        with open(file_path, "r") as src, open(target_path, "w") as dst:
            dst.write(src.read())

        show_success_message(f"Example added as {target_path.name}")

    elif action == "from-pr":
        pr_number = prompt("Enter PR link: ")
        file_extension = prompt(
            "Enter file extension for examples (default: java): ", default="java"
        )

        async def generate_examples():
            pr_details = await get_pr_details(pr_number)
            examples = await extract_example_patterns(pr_number, pr_details)

            if not examples:
                print(
                    "Warning: No example patterns were extracted. Creating a simple example..."
                )

                examples = [
                    (
                        "// Old version\npublic class Example {\n    // TODO: Add your code here\n}",
                        "// New version\npublic class Example {\n    // TODO: Migrated code here\n}",
                    )
                ]

            await save_examples(examples, examples_dir, file_extension)
            return len(examples)

        num_examples, error = run_async_with_progress(
            "Generating examples from PR...", generate_examples
        )

        if error:
            show_error_message("Error generating examples", error)
        elif num_examples > 0:
            show_success_message(
                f"Generated {num_examples} example pairs from PR #{pr_number}"
            )
        else:
            show_warning_message(f"No examples were generated from PR #{pr_number}")
            console.print("You may need to create examples manually.")

    elif action == "setup":
        ref = prompt("Enter git ref: ")
        pattern = prompt("Enter file pattern (optional): ", default="")

        result, error = run_with_progress(
            "Setting up examples from git history...",
            setup_examples,
            ref,
            str(examples_dir),
            pattern if pattern else None,
        )

        if error:
            show_error_message("Error setting up examples", error)
        else:
            show_success_message(f"Examples set up from git ref {ref}")


def manage_evals(project_dir: Path):
    """Manage evaluation test cases."""
    evals_dir = project_dir / "evals"
    if not evals_dir.exists():
        evals_dir.mkdir(exist_ok=True)
        show_success_message(f"Created evals directory at {evals_dir}")

    action = console_radiolist(
        title="Evaluations Management",
        text="What would you like to do?",
        values=[
            ("list", "List existing evaluations"),
            ("from-pr", "Generate evaluations from a PR"),
            ("from-migration", "Generate evaluation from a recent migration"),
        ],
    )

    if not action:
        return

    if action == "list":
        eval_dirs = [d for d in evals_dir.iterdir() if d.is_dir()]

        if not eval_dirs:
            console.print("No evaluations found.")
            return

        table = Table(title="Evaluations")
        table.add_column("Name", style="cyan")
        table.add_column("Files", style="green")
        table.add_column("Created", style="blue")

        for eval_dir in eval_dirs:
            source_dir = eval_dir / "source"
            files = list(source_dir.glob("*")) if source_dir.exists() else []
            created = datetime.fromtimestamp(eval_dir.stat().st_ctime).strftime(
                "%Y-%m-%d %H:%M"
            )
            table.add_row(eval_dir.name, str(len(files)), created)

        console.print(table)

    elif action == "from-pr":
        pr_number = prompt("Enter PR link: ")

        result, error = run_async_with_progress(
            "Generating evaluations from PR...",
            generate_eval_from_pr,
            pr_number,
            str(project_dir),
        )

        if error:
            show_error_message("Error generating evaluations", error)
            return error
        else:
            show_success_message(
                f"Generated evaluations from PR #{pr_number} at {result}"
            )

    elif action == "from-migration":
        console.print("This feature is not yet implemented.")


def manage_system_prompt(project_dir: Path):
    """View or edit the system prompt."""
    system_prompt_path = project_dir / SYSTEM_PROMPT_FILE

    if not system_prompt_path.exists():
        show_error_message(f"System prompt file {system_prompt_path} does not exist.")
        return

    action = console_radiolist(
        title="System Prompt",
        text="What would you like to do?",
        values=[
            ("view", "View the current system prompt"),
            ("edit", "Edit the system prompt"),
            ("generate", "Generate a new system prompt"),
        ],
    )

    if not action:
        return

    if action == "view":
        content = system_prompt_path.read_text()
        console.print(
            Panel(Markdown(content), title="System Prompt", border_style="blue")
        )

    elif action == "edit":
        content = system_prompt_path.read_text()

        with tempfile.NamedTemporaryFile(suffix=".md", mode="w", delete=False) as tmp:
            tmp.write(content)
            tmp_path = tmp.name

        editor = os.environ.get("EDITOR", "nano")
        subprocess.run([editor, tmp_path])

        with open(tmp_path, "r") as f:
            new_content = f.read()

        os.unlink(tmp_path)

        if new_content != content:
            system_prompt_path.write_text(new_content)
            show_success_message("System prompt updated.")
        else:
            console.print("System prompt not changed.")

    elif action == "generate":
        method = console_radiolist(
            title="Generation Method",
            text="How would you like to generate the system prompt?",
            values=[
                ("pr", "From a PR"),
                ("description", "From a description"),
            ],
        )

        if not method:
            return

        if method == "pr":
            pr_number = prompt("Enter PR link: ")
            description = prompt("Enter a brief description of the migration task: ")

            async def generate_prompt():
                pr_details = await get_pr_details(pr_number)
                return await generate_system_prompt(pr_details, description)

            system_prompt, error = run_async_with_progress(
                "Generating system prompt...", generate_prompt
            )

            if error:
                show_error_message("Error generating system prompt", error)
                return

        elif method == "description":
            description = prompt("Enter a detailed description of the migration task: ")

            async def generate_prompt():
                client = DefaultClient()
                system_prompt_for_generation = """You are an expert at creating system prompts for code migration tasks.
Create a clear, concise system prompt that will guide an AI to perform code migrations based on the description provided."""

                user_prompt = f"""
I need to create a system prompt for a code migration task. Here's the description:

{description}

Create a system prompt that clearly explains:
1. The migration goal
2. Key patterns to look for
3. How to transform the code
4. Any important constraints or considerations

The system prompt should be concise but comprehensive, focusing on the migration patterns.
"""

                return await client.generate_text(
                    system_prompt_for_generation, user_prompt
                )

            system_prompt, error = run_async_with_progress(
                "Generating system prompt...", generate_prompt
            )

            if error:
                show_error_message("Error generating system prompt", error)
                return

        console.print(
            Panel(
                Markdown(system_prompt),
                title="Generated System Prompt",
                border_style="blue",
            )
        )

        if console_yes_no(
            title="Save System Prompt",
            text="Do you want to save this system prompt?",
        ):
            system_prompt_path.write_text(system_prompt)
            show_success_message(f"System prompt saved to {system_prompt_path}")


@click.group()
def cli():
    """AI Migrate - Code Migration Tool."""
    print_header()


@cli.command()
@click.option(
    "--interactive/--no-interactive", default=True, help="Run in interactive mode"
)
def init(interactive):
    """Initialize a new migration project."""
    if not interactive:
        click.echo("Please use the interactive mode for better experience.")
        return

    project_name = prompt("Enter project name: ")
    default_path = Path.cwd()
    if not default_path.name == "projects":
        default_path = default_path / "projects"
    default_path = default_path / project_name
    project_path = Path(
        prompt(
            "Enter project directory path: ",
            completer=PathCompleter(),
            default=str(default_path),
        ).strip()
    )

    if project_path.exists() and any(project_path.iterdir()):
        if not console_yes_no(
            title="Directory not empty",
            text=f"The directory {project_path} already exists and is not empty. Continue anyway?",
        ):
            return

    use_pr = console_yes_no(
        title="PR-Based Initialization",
        text="Do you want to initialize the project based on a GitHub PR?",
    )

    generate_evals = False

    if use_pr:
        pr_number = prompt("Enter PR link: ")
        description = prompt("Enter a brief description of the migration task: ")
        file_extension = prompt(
            "Enter file extension for examples (default: java): ", default="java"
        )

        generate_evals = console_yes_no(
            title="Generate Evaluations",
            text="Do you want to generate evaluation files from this PR?",
        )

        result, error = run_async_with_progress(
            "Initializing project from PR...",
            setup_project_from_pr,
            pr_number,
            str(project_path),
            description,
            file_extension,
        )

        if error:
            show_error_message(f"Error during project initialization: {error}")
            show_warning_message(
                "Project was created with minimal files. Some manual editing may be required."
            )
        else:
            show_success_message(f"Project initialized successfully at {project_path}")

            if generate_evals and pr_number:
                result, error = run_async_with_progress(
                    "Generating evaluation files from PR...",
                    generate_eval_from_pr,
                    pr_number,
                    str(project_path),
                )

                if error:
                    show_error_message(f"Error generating evaluation files: {error}")
                    show_warning_message(
                        "You may need to create evaluation files manually."
                    )
                else:
                    show_success_message("Evaluation files generated successfully")
    else:
        project_path.mkdir(parents=True, exist_ok=True)
        examples_dir = project_path / "examples"
        examples_dir.mkdir(exist_ok=True)
        evals_dir = project_path / "evals"
        evals_dir.mkdir(exist_ok=True)

        system_prompt = project_path / SYSTEM_PROMPT_FILE
        system_prompt.write_text(SYSTEM_MESSAGE)

        show_success_message(f"Project initialized successfully at {project_path}")

    console.print("\nNext steps:")
    console.print("1. Review the generated system prompt")
    console.print("2. Review the generated examples")
    console.print(
        f"\nTo set this as your default project: [bold]export AI_MIGRATE_PROJECT_DIR={project_path}[/bold]"
    )


def project_dir_option(f):
    def project_dir_validate(ctx, param, project_dir):
        if "project_dir" in ctx.params:
            return ctx.params["project_dir"]

        if "--project" in param.opts:
            project_dir = get_project_dir(project_dir)
            return project_dir

        if not project_dir:
            project_dir = os.environ.get("AI_MIGRATE_PROJECT_DIR")

        if not project_dir and Path(".ai-migrate").exists():
            try:
                project_dir = json.loads(Path(".ai-migrate").read_text())["project_dir"]
            except (json.JSONDecodeError, KeyError):
                pass

        if not project_dir:
            console.print("[yellow]No project directory specified.[/yellow]")
            console.print(
                "Please specify a project directory using --project-dir or set AI_MIGRATE_PROJECT_DIR environment variable."
            )
            raise click.MissingParameter("project_dir")

        project_dir = Path(project_dir).expanduser().absolute()

        if not project_dir.exists():
            raise click.BadParameter(f"Project directory {project_dir} does not exist.")

        return project_dir

    return click.option(
        "--project",
        "project_dir",
        help="A pre-configured project",
        callback=project_dir_validate,
    )(
        click.option(
            "--project-dir",
            "project_dir",
            help="Path to the project directory",
            callback=project_dir_validate,
        )(f)
    )


@cli.command()
@click.argument("file_paths", nargs=-1)
@project_dir_option
@click.option(
    "--manage",
    type=click.Choice(["examples", "system-prompt", "evals"]),
    help="Manage examples, system prompt, or evaluations",
)
@click.option("--manifest-file", help="Path to the manifest file")
@click.option(
    "--rerun-passed/--only-failed",
    default=False,
    help="Re-run migrations that have already passed",
)
@click.option(
    "--max-workers", default=8, type=int, help="Maximum number of parallel workers"
)
@click.option(
    "--local-worktrees/--no-local-worktrees",
    default=False,
    help="Create worktrees alongside the git repo",
)
@click.option(
    "--llm-fakes",
    default=None,
    type=str,
    help="Use fake LLM responses for testing",
)
@click.option(
    "--dont-create-evals",
    is_flag=True,
    help="Don't automatically create evaluations after successful migrations",
)
def migrate(
    file_paths,
    project_dir,
    manage,
    manifest_file,
    rerun_passed,
    max_workers,
    local_worktrees,
    llm_fakes,
    dont_create_evals,
):
    """Migrate one or more files or manage project resources.

    If --manage is specified, you can manage examples, system prompt, or evaluations instead of migrating files.
    """
    console.print(f"Using project: [bold cyan]{project_dir}[/bold cyan]")

    if manage:
        if manage == "examples":
            manage_examples(project_dir)
            return
        elif manage == "system-prompt":
            manage_system_prompt(project_dir)
            return
        elif manage == "evals":
            manage_evals(project_dir)
            return

    from .projects import run as projects_run

    if not file_paths and not manifest_file:
        if (make_manifest_script := (Path(project_dir) / "make_manifest.py")).exists():
            console.print("Found make_manifest.py in project directory. Running...")
            try:
                manifest_json = subprocess.run(
                    [sys.executable, make_manifest_script],
                    capture_output=True,
                    check=True,
                ).stdout.decode()
            except subprocess.CalledProcessError as e:
                show_error_message("Error running make_manifest.py", e.stderr.decode())
                raise
            dt = datetime.now().strftime("%Y%m%d-%H%M%S")
            manifest_file = f"manifest-{dt}.json"
            with open(manifest_file, "w") as f:
                f.write(manifest_json)
            print("Manifest file created:", manifest_file)
        elif not IS_INTERACTIVE:
            console.print(
                "[yellow]No files specified and not in interactive terminal. Please specify files using command line arguments.[/yellow]"
            )
            return
        else:
            file_path = prompt(
                "Enter file path to migrate: ", completer=PathCompleter()
            )
            file_paths = [file_path]

    run_id = datetime.now().strftime("%Y%m%d-%H%M%S")
    logs_dir = Path("ai-migrator-logs") / run_id
    result = asyncio.run(
        projects_run(
            str(project_dir),
            logs_dir,
            manifest_file,
            file_paths,
            not rerun_passed,
            max_workers,
            local_worktrees,
            resume=True,
            llm_fakes=llm_fakes,
            dont_create_evals=dont_create_evals,
        )
    )

    has_failures = any("fail" in fg.result for fg in result)
    has_pre_verify_failures = any(fg.result == "pre-verify-fail" for fg in result)
    if has_failures:
        show_error_message("Migration failed: some files did not pass verification")
        console.print(f"Logs are available at {logs_dir}")
        console.print(f"View them with `ai-migrate logs {run_id}`")
        if has_pre_verify_failures:
            console.print("Some files failed pre-verification.")
            logs()

        sys.exit(1)


@cli.command()
@click.option("--project-dir", help="Path to the project directory")
@click.option("--manifest")
def status(
    project_dir,
    manifest,
):
    """Show the status of migration projects."""

    from .projects import status as projects_status

    result, error = run_with_progress("Getting migration status...", projects_status)

    if error:
        show_error_message("Failed to get migration status", error)


@cli.command()
@click.argument("file_path")
@project_dir_option
def checkout(file_path, project_dir):
    """Check out the branch for a file to try manual fixes on a failed migration attempt."""

    from .projects import checkout_failed

    result, error = run_with_progress(
        f"Checking out branch for {file_path}...", checkout_failed, file_path
    )

    if error:
        show_error_message(f"Failed to checkout branch for {file_path}", error)
    else:
        show_success_message(f"Successfully checked out branch for {file_path}")


@cli.command()
@click.argument("manifest")
def merge_branches(manifest):
    """Merge the changes from the migrator branches."""
    from .merge_migrator_changes import merge

    result, error = run_with_progress(
        "Merging changes from migrator branches...", merge, manifest
    )

    if error:
        show_error_message("Failed to merge changes", error)
    else:
        show_success_message("Successfully merged changes from migrator branches")


@cli.command
@click.argument("files", nargs=-1)
@project_dir_option
def verify(project_dir, files):
    """Run the verification step over a file"""
    from .projects import verify

    verify(project_dir, files, manifest_file=None)


@cli.command()
@click.argument("run_id", default="latest")
def logs(run_id):
    """Show the logs for a given run."""
    if run_id == "latest":
        candidates = sorted(
            [
                logs_dir.name
                for logs_dir in Path("ai-migrator-logs").iterdir()
                if logs_dir.is_dir() and re.match(r"^\d{8}-\d{6}$", logs_dir.name)
            ]
        )
        if not candidates:
            console.print("[red]No logs found.[/red]")
            return
        run_id = candidates[-1]
        print(f"Using latest run: {run_id}")

    logs_dir = Path("ai-migrator-logs") / run_id

    for log_file in logs_dir.glob("*.log"):
        print(f"{log_file}:")
        with open(log_file) as f:
            print(f.read())
        print()


@cli.command()
@project_dir_option
@click.argument("pr_num")
def add_examples_from_pr(pr_num, project_dir):
    setup_from_pr(pr_num, Path(project_dir) / "examples")


@cli.command()
@click.argument("manifest")
@click.option("--force", default=False, is_flag=True)
def cleanup_git_branches(manifest, force):
    """Clean up git worktrees and branches created by the migrator."""
    branches = {
        branch
        for _, branch, *_ in get_branches(
            Manifest.model_validate_json(Path(manifest).read_text())
        )
    }

    subprocess.run(["git", "worktree", "prune"], check=True)
    for worktree, branch, prunable in get_worktrees():
        if branch in branches:
            if prunable or force:
                subprocess.run(["git", "worktree", "remove", "--force", worktree])
            else:
                console.print(
                    f"[yellow]Worktree {worktree} for branch {branch} is not prunable. Use --force to remove it.[/yellow]"
                )
                branches.remove(branch)

    for branch in branches:
        assert branch.startswith("ai-migrator/")
        subprocess.run(["git", "branch", "-D" if force else "-d", branch])


@cli.command()
@project_dir_option
@click.argument("script")
def script(script, project_dir):
    """Run a script from the project directory"""
    script_path = Path(project_dir) / script
    if (py_script := script_path.with_suffix(".py")).exists():
        command = [sys.executable, py_script]
    elif (sh_script := script_path.with_suffix(".sh")).exists():
        command = [sh_script]
    else:
        show_error_message(f"Script {script} not found in project directory.")
        return

    console.print(f"Running script: [bold cyan]{script_path}[/bold cyan]")
    subprocess.run(command, check=True)


@cli.command()
@project_dir_option
def current_project(project_dir):
    """Show the project directory."""
    console.print(f"Project directory: [bold cyan]{project_dir}[/bold cyan]")


def main():
    """Entry point for the interactive CLI."""
    try:
        cli()
    except KeyboardInterrupt:
        console.print("\n[yellow]Operation cancelled by user.[/yellow]")
        sys.exit(1)
    except Exception as e:
        error_msg = str(e).replace("[", "\\[").replace("]", "\\]")
        console.print(f"\n[red]Error:[/red] {error_msg}")
        console.print_exception()
        sys.exit(1)


if __name__ == "__main__":
    main()



from importlib.metadata import entry_points

from pathlib import Path
from typing import Iterable


class ProjectsRoot:
    def __init__(self, root_dir: Path | str):
        self.root_dir = Path(root_dir)

    def list_projects(self) -> Iterable[Path]:
        if not self.root_dir.exists() or not self.root_dir.is_dir():
            return []
        for p in self.root_dir.iterdir():
            if p.is_dir():
                yield p


_ROOTS = [
    ProjectsRoot(Path("~/ai-migrator-projects").expanduser()),
]
_EXTRA_ROOTS_LOADED = False


def _load_roots():
    global _EXTRA_ROOTS_LOADED
    for ep in entry_points(group="ai_migrate").select(name="projects_root"):
        roots = ep.load()
        if hasattr(roots, "__iter__"):
            _ROOTS.extend(roots)
        else:
            _ROOTS.append(roots)
    _EXTRA_ROOTS_LOADED = True


def get_project_dir(name: str) -> Path:
    if not _EXTRA_ROOTS_LOADED:
        _load_roots()
    for root in _ROOTS:
        for project in root.list_projects():
            if project.name == name:
                return project
    raise FileNotFoundError(
        f"Project {name} not found in project roots [{', '.join(str(r.root_dir) for r in _ROOTS)}]"
    )



import subprocess
from pathlib import Path

from ai_migrate.git import get_branches
from ai_migrate.manifest import Manifest


def merge(manifest_file: str):
    manifest = Manifest.model_validate_json(Path(manifest_file).read_text())
    matching_branches = [
        (branch, status) for _, branch, status, _ in get_branches(manifest)
    ]
    mergeable = [b for b, status in matching_branches if status == "pass"]
    print("Merging:")
    print("\n".join(mergeable))
    print("Ignoring:\n")
    print(
        "\n".join(
            f"{b} ({status})" for b, status in matching_branches if b not in mergeable
        ),
    )

    if not matching_branches:
        raise ValueError("No matching branches")

    subprocess.run(["git", "merge", "--squash", *mergeable], check=True)


if __name__ == "__main__":
    merge()



import pytest
from ai_migrate.resolve_symbols import contains_definition


@pytest.mark.parametrize(
    "line, symbol, expected",
    [
        # Class definitions
        ("public class MyClass", "MyClass", True),
        ("class MyClass", "MyClass", True),
        ("public class MyOtherClass", "MyClass", False),
        # Generic class definitions
        ("data class EndFlowAction<out TString>", "EndFlowAction", True),
        ("class GenericClass<T>", "GenericClass", True),
        ("class GenericClass<T>", "OtherClass", False),
        # Interface definitions
        ("interface MyInterface", "MyInterface", True),
        ("interface MyOtherInterface", "MyInterface", False),
        # Method definitions
        ("public void myMethod()", "myMethod", True),
        ("int getValue()", "getValue", True),
        ("void anotherMethod()", "myMethod", False),
        # Mixed modifiers and keywords
        ("private final class MixedClass", "MixedClass", True),
        ("public static void main(String[] args)", "main", True),
        ("enum MyEnum { VALUE1, VALUE2 }", "MyEnum", True),
        ("enum MyEnum { VALUE1, VALUE2 }", "OtherEnum", False),
        # Edge cases
        ("data class Empty<out T>()", "Empty", True),
        ("data class NotEmpty<out T>()", "Other", False),
        ("class MyClass /* comment */", "MyClass", True),
        ("public static String myStaticMethod()", "myStaticMethod", True),
    ],
)
def test_contains_definition(line, symbol, expected):
    assert contains_definition(line, symbol) == expected


if __name__ == "__main__":
    pytest.main()



import os
import shutil
import subprocess
import sys
import tempfile
import time
import logging
import argparse
from pathlib import Path

from ai_migrate.git_identity import environment_variables
from ai_migrate.manifest import Manifest

SCRIPT_DIR = Path(__file__).parent.parent.absolute()
AI_MIGRATE_PROJECT_DIR = SCRIPT_DIR.parent
EVALS_DIR = SCRIPT_DIR / "source"

# Configure logging
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(levelname)s - %(message)s",
    handlers=[logging.StreamHandler(sys.stdout)],
)
logger = logging.getLogger("ai_migrate.run_eval")


class Workspace:
    temp_dir: Path

    def __init__(self, verbose=False):
        self.temp_dir = Path(tempfile.mkdtemp())
        self.working_dir = self.temp_dir
        self.verbose = verbose
        logger.debug(f"Created workspace at {self.temp_dir}")

    def sh(self, command, pipe_out=False, env=None):
        if env is None:
            env = {}

        logger.log(
            logging.INFO if self.verbose else logging.DEBUG,
            f"Running command: {command} in {self.working_dir}",
        )
        stderr_redirect = stdout_redirect = (
            None if pipe_out or self.verbose else subprocess.PIPE
        )

        try:
            result = subprocess.run(
                command.split(),
                check=True,
                cwd=self.working_dir,
                stdout=stdout_redirect,
                stderr=stderr_redirect,
                env={**os.environ, **env},
            )
        except subprocess.CalledProcessError as e:
            logger.error(f"Command failed with return code {e.returncode}")
            logger.error("stdout:")
            logger.error(e.stdout)
            logger.error("stderr:")
            logger.error(e.stderr)
            raise

        logger.debug(f"Command completed with return code: {result.returncode}")
        return result

    def pushd(self, directory):
        logger.debug(
            f"Changing directory from {self.working_dir} to {self.working_dir / directory}"
        )
        self.working_dir = self.working_dir / directory

    def popd(self):
        if self.working_dir == self.temp_dir:
            raise ValueError("Cannot popd past the initial directory")
        old_dir = self.working_dir
        self.working_dir = self.working_dir.parent
        logger.debug(f"Changing directory back from {old_dir} to {self.working_dir}")

    def cleanup(self):
        logger.debug(f"Cleaning up workspace at {self.temp_dir}")
        shutil.rmtree(self.temp_dir)
        print("done cleaning up")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            logger.error(f"Exception in workspace: {exc_type.__name__}: {exc_val}")
        self.cleanup()


def run_project_eval(project, verbose=False):
    migrate_cmd = f"uv run --project {AI_MIGRATE_PROJECT_DIR} ai-migrate"
    with Workspace(verbose=verbose) as ws:
        logger.info(f"Running project eval for '{project}' in workspace {ws.temp_dir}")
        if verbose:
            logger.info("Verbose mode enabled - showing detailed output")

        start_time = time.time()
        project_dir = ws.temp_dir / "migrate-project"

        logger.info(
            f"Copying project files from {AI_MIGRATE_PROJECT_DIR / 'projects' / project} to {project_dir}"
        )
        shutil.copytree(AI_MIGRATE_PROJECT_DIR / "projects" / project, project_dir)

        eval_count = 0
        pass_count = 0
        fail_count = 0

        for directory in (project_dir / "evals").iterdir():
            eval_count += 1
            manifest_file = directory / "manifest.json"
            logger.info(f"Processing eval: {project}/{directory.name}")
            logger.info(f"Reading manifest file: {manifest_file}")

            manifest = Manifest.model_validate_json(manifest_file.read_text())
            logger.info(f"Manifest loaded with {len(manifest.files)} file entries")

            if manifest.eval_target_repo_remote and manifest.eval_target_repo_ref:
                logger.info(
                    f"Cloning target repo: {manifest.eval_target_repo_remote} @ {manifest.eval_target_repo_ref}"
                )
                ws.sh(f"git clone --depth=1 {manifest.eval_target_repo_remote} repo")
                ws.pushd("repo")
                ws.sh(f"git fetch --depth 1 origin {manifest.eval_target_repo_ref}")
                ws.sh(f"git checkout {manifest.eval_target_repo_ref}")
            else:
                logger.info(f"Using local source files from: {directory / 'source'}")
                shutil.copytree(directory / "source", ws.working_dir / "repo")
                ws.pushd("repo")
                ws.sh("git init")
                ws.sh("git add .")
                ws.sh("git commit -m initialcommit", env=environment_variables())

            try:
                logger.info(f"Running migration with manifest: {manifest_file}")
                command = [
                    migrate_cmd,
                    "migrate",
                    "--manifest-file",
                    str(manifest_file),
                    "--local-worktrees",
                    "--project-dir",
                    str(project_dir),
                ]
                if (fakes_directory := directory / "llm-fakes").exists():
                    command.extend(["--llm-fakes", str(fakes_directory)])
                ws.sh(
                    " ".join(command),
                    env={"AI_MIGRATE_PROJECT_DIR": project_dir},
                    pipe_out=True,
                )
                result = "PASSED"
                pass_count += 1
                logger.info(f"✅ Migration successful for {project}/{directory.name}")
            except subprocess.CalledProcessError:
                result = "FAILED"
                fail_count += 1
                logger.error(
                    f"❌ Error running migration for {project}/{directory.name}"
                )

                # Always show logs for failed migrations, regardless of verbose mode
                log_files = list((ws.working_dir / "ai-migrator-logs").glob("**/*.log"))
                logger.error(f"Found {len(log_files)} log files:")

                for log_file in log_files:
                    logger.error(f"Contents of {log_file}:")
                    log_content = log_file.read_text()
                    for line in log_content.splitlines():
                        logger.error(f"  | {line}")
                    logger.error("")

            print(f"{project}/{directory.name}", result)

            # Return to the parent directory for the next eval
            ws.popd()

        elapsed_time = time.time() - start_time
        logger.info(f"Completed {eval_count} evals in {elapsed_time:.2f}s")
        logger.info(f"Results: {pass_count} passed, {fail_count} failed")
        print(f"Done in {elapsed_time:.2f}s")

        return {
            "project": project,
            "total": eval_count,
            "passed": pass_count,
            "failed": fail_count,
            "time": elapsed_time,
        }


def main():
    parser = argparse.ArgumentParser(
        description="Run evaluations for AI migration projects"
    )
    parser.add_argument(
        "projects",
        nargs="*",
        help="Project names to evaluate. If none provided, all projects will be evaluated.",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Enable verbose output"
    )
    parser.add_argument(
        "--log-level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="INFO",
        help="Set the logging level",
    )
    parser.add_argument("--log-file", help="Log to a file in addition to console")

    args = parser.parse_args()

    # Configure logging based on arguments
    log_level = getattr(logging, args.log_level)
    logger.setLevel(log_level)

    # Add file handler if log file is specified
    if args.log_file:
        file_handler = logging.FileHandler(args.log_file)
        file_handler.setFormatter(
            logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
        )
        logger.addHandler(file_handler)
        logger.info(f"Logging to file: {args.log_file}")

    projects = args.projects

    if not projects:
        logger.info("No projects specified, discovering available projects...")
        projects_dir = Path(AI_MIGRATE_PROJECT_DIR / "projects")
        projects = []

        for directory in projects_dir.iterdir():
            if directory.is_dir() and (directory / "evals").is_dir():
                projects.append(directory.name)
                logger.info(f"Found project: {directory.name}")

    logger.info(
        f"Running evaluations for {len(projects)} projects: {', '.join(projects)}"
    )

    total_results = {
        "total_projects": len(projects),
        "total_evals": 0,
        "passed_evals": 0,
        "failed_evals": 0,
        "projects": [],
    }

    for project in projects:
        logger.info(f"Starting evaluation for project '{project}'")
        print(f"Running eval for project '{project}'")

        result = run_project_eval(project, verbose=args.verbose)
        total_results["total_evals"] += result["total"]
        total_results["passed_evals"] += result["passed"]
        total_results["failed_evals"] += result["failed"]
        total_results["projects"].append(result)

    # Print summary
    logger.info("=" * 50)
    logger.info("EVALUATION SUMMARY")
    logger.info("=" * 50)
    logger.info(f"Total projects evaluated: {total_results['total_projects']}")
    logger.info(f"Total evals run: {total_results['total_evals']}")
    logger.info(f"Passed evals: {total_results['passed_evals']}")
    logger.info(f"Failed evals: {total_results['failed_evals']}")

    if total_results["failed_evals"] > 0:
        logger.warning("Some evaluations failed. Check the logs for details.")
        return 1
    else:
        logger.info("All evaluations passed successfully!")
        return 0


if __name__ == "__main__":
    sys.exit(main())



from pathlib import Path

import pytest

from ai_migrate.migrate import (
    FileContent,
    migrate_prompt,
    extract_code_blocks,
    CodeResponseResult,
    CodeBlock,
    MigrationExample,
    read_file_pairs_from,
)


@pytest.fixture
def examples_dir():
    return Path(__file__).parent / "test" / "examples"


def test_read_file_pairs_from(examples_dir):
    # Convert iterator to list for testing
    examples = list(read_file_pairs_from(examples_dir))

    # Sort examples by name for consistent testing
    examples.sort(key=lambda x: x.name if x.name else "")

    # We expect 2 examples: one directory pair and one file pair
    assert len(examples) == 2

    # Test directory pair example
    dir_example = examples[0]
    assert dir_example.name == "example1"

    # Sort files by name for consistent testing
    dir_example.old_files.sort(key=lambda x: x.name)
    dir_example.new_files.sort(key=lambda x: x.name)

    # Check old files
    assert len(dir_example.old_files) == 2
    assert dir_example.old_files[0].name == "src/file1.py"
    assert dir_example.old_files[1].name == "src/file2.py"
    assert (
        dir_example.old_files[0].content
        == """def old_function():
    print("This is the old version of file1")\n"""
    )
    assert (
        dir_example.old_files[1].content
        == """class OldClass:
    def method(self):
        return "old implementation"\n"""
    )

    # Check new files
    assert len(dir_example.new_files) == 2
    assert dir_example.new_files[0].name == "src/file1.py"
    assert dir_example.new_files[1].name == "src/file2.py"
    assert (
        dir_example.new_files[0].content
        == """def new_function():
    print("This is the new version of file1")\n"""
    )
    assert (
        dir_example.new_files[1].content
        == """class NewClass:
    def method(self):
        return "new implementation"\n"""
    )

    # Test single file pair example
    file_example = examples[1]
    assert file_example.name == "single"

    # Check old file
    assert len(file_example.old_files) == 1
    assert file_example.old_files[0].name == "single.py"
    assert (
        file_example.old_files[0].content
        == """def standalone_old():
    return "old standalone function"\n"""
    )

    # Check new file
    assert len(file_example.new_files) == 1
    assert file_example.new_files[0].name == "single.py"
    assert (
        file_example.new_files[0].content
        == """def standalone_new():
    return "new standalone function"\n"""
    )


def test_examples_prompt():
    example = MigrationExample(
        name="main",
        old_files=[
            FileContent(
                name="main.kt",
                content="""
fun main() {
    apiv1("Hello, world!")
}
""",
            )
        ],
        new_files=[
            FileContent(
                name="main.kt",
                content="""
fun main() {
    apiv2("Hello, world!")
}
""",
            )
        ],
    )
    messages = migrate_prompt(example)
    assert messages == [
        {
            "role": "user",
            "content": """Migrate this code to the new format:

### `main.kt`
```kotlin
fun main() {
    apiv1("Hello, world!")
}
```. Return the full content for all files mentioned, don't leave anything out. You can rename a file if necessary.""",
        },
        {
            "role": "assistant",
            "content": """Here's the migrated code:
### `main.kt`
```kotlin
fun main() {
    apiv2("Hello, world!")
}
```""",
        },
    ]


def test_no_new_code_prompt():
    example = MigrationExample(
        name="main",
        old_files=[
            FileContent(
                name="main.kt",
                content="""
fun main() {
    apiv1("Hello, world!")
}
""",
            )
        ],
        new_files=[],
    )
    messages = migrate_prompt(example)
    assert messages == [
        {
            "role": "user",
            "content": """Migrate this code to the new format:

### `main.kt`
```kotlin
fun main() {
    apiv1("Hello, world!")
}
```. Return the full content for all files mentioned, don't leave anything out. You can rename a file if necessary.""",
        }
    ]


def test_multifile_prompt():
    example = MigrationExample(
        name="multifile",
        old_files=[
            FileContent(
                name="main.kt",
                content="""
fun main() {
    hello("Hello, world!")
}
""",
            ),
            FileContent(
                name="hello.kt",
                content="""
fun hello(message: String) {
    println(message)
}
""",
            ),
        ],
        new_files=[
            FileContent(
                name="main.kt",
                content="""
fun main() {
    hello("Hello, world!", "again!")
}
""",
            ),
            FileContent(
                name="hello.kt",
                content="""
fun hello(message: String, message2: String) {
    println(message)
    println(message2)
}
""",
            ),
        ],
    )
    messages = migrate_prompt(example)
    assert messages == [
        {
            "role": "user",
            "content": """Migrate this code to the new format:

### `main.kt`
```kotlin
fun main() {
    hello("Hello, world!")
}
```

### `hello.kt`
```kotlin
fun hello(message: String) {
    println(message)
}
```. Return the full content for all files mentioned, don't leave anything out. You can rename a file if necessary.""",
        },
        {
            "role": "assistant",
            "content": """Here's the migrated code:
### `main.kt`
```kotlin
fun main() {
    hello("Hello, world!", "again!")
}
```

### `hello.kt`
```kotlin
fun hello(message: String, message2: String) {
    println(message)
    println(message2)
}
```""",
        },
    ]


def test_split_code_blocks():
    markdown = """Here's some code:
### `file1.py`
```python
def hello():
    pass
```
try it out. Some more:
```
uv run file1.py
```
"""
    result = extract_code_blocks(markdown)
    assert result == CodeResponseResult(
        code_blocks=[
            CodeBlock(filename="file1.py", code="def hello():\n    pass"),
            CodeBlock(filename=None, code="uv run file1.py"),
        ],
        other_text="Here's some code:\n<code>\ntry it out. Some more:\n<code>",
    )



import asyncio
import sys
import itertools
from typing import Dict
import shutil


class StatusBar:
    def __init__(self, name: str = "Task"):
        self.name = name
        self.status = None  # None = in progress, "passed" or "failed"
        self.message = ""  # Status message to display
        self.spinner_chars = "⠋⠙⠹⠸⠼⠴⠦⠧⠇⠏"
        self.spinner_iter = itertools.cycle(self.spinner_chars)

    def render(self) -> str:
        # Get terminal width, fallback to 80 if not available
        terminal_width = shutil.get_terminal_size((80, 20)).columns
        # Get the status symbol part
        if self.status == "passed":
            status_symbol = "✓"
        elif self.status == "failed":
            status_symbol = "✗"
        else:
            status_symbol = next(self.spinner_iter)

        # Create the right-aligned part (status symbol and optional message)
        right_part = status_symbol
        if self.message:
            right_part = f"{status_symbol} - {self.message}"

        # Combine with name, padding the right part to align with terminal width
        name_part = f"{self.name}: "
        padding = terminal_width - len(name_part) - len(right_part)
        return f"\r{name_part}{' ' * max(0, padding)}{right_part}"


class StatusManager:
    def __init__(self):
        self.bars: Dict[str, StatusBar] = {}
        self.lock = asyncio.Lock()
        self.is_terminal = sys.stdout.isatty()
        self._running = True
        self._render_task = None
        self._last_render_lines = 0  # Will be 0 for first render

    async def add_status(self, name: str) -> StatusBar:
        async with self.lock:
            bar = StatusBar(name)
            self.bars[name] = bar
            if self._render_task is None:
                self._render_task = asyncio.create_task(self._render_loop())
            await self.render()
            return bar

    async def _render_loop(self):
        while self._running:
            await self.render()
            await asyncio.sleep(0.1)

    async def render(self):
        if not self.is_terminal:
            return

        # Move up by the number of lines we last rendered
        sys.stdout.write("\033[2K\033[A" * self._last_render_lines)
        sys.stdout.flush()

        # Count how many lines we're rendering
        self._last_render_lines = len(self.bars)
        bars = list(self.bars.values())
        bars = sorted(bars, key=lambda bar: (bar.status or "~", bar.name))
        for bar in bars:
            print(bar.render())

    async def mark_passed(self, name: str):
        async with self.lock:
            if name in self.bars:
                self.bars[name].status = "passed"
                await self.render()

    async def mark_failed(self, name: str):
        async with self.lock:
            if name in self.bars:
                self.bars[name].status = "failed"
                await self.render()

    async def stop(self):
        self._running = False
        if self._render_task:
            self._render_task.cancel()
            try:
                await self._render_task
            except asyncio.CancelledError:
                pass

    async def update_message(self, name: str, message: str):
        async with self.lock:
            if name in self.bars:
                self.bars[name].message = message
                await self.render()



import sys
import subprocess

filename = sys.argv[1]
out = (
    subprocess.run(["python", filename], check=True, capture_output=True)
    .stdout.decode()
    .strip()
)
assert out == "fizzbuzz\nfizzbuzz\nfizzbuzz"



import asyncio
import contextvars
import os
import shutil
import sys
import tempfile
from dataclasses import dataclass
from pathlib import Path
import subprocess
import json
from typing import Any, Iterable, Optional

from ai_migrate.llm_providers import DefaultClient
from .fake_llm_client import FakeLLMClient
from .git_identity import environment_variables
from .manifest import FileGroup, FileEntry, Manifest
from .resolve_symbols import get_symbol_definition
from .eval_generator import generate_eval_from_migration

FN_LOOKUP_SYMBOL = "lookup_symbol_definition"

LOOKUP_SYMBOL_PROMPT = (
    "Do not include any explanations or comments outside the code block.\n"
    "Do not assume functionality of classes or functions that you haven't seen in the examples. "
    f"Instead use the provided tool '{FN_LOOKUP_SYMBOL}' to look up the definition of a symbol.\n"
    "You want to produce code that you can guarantee to compile, so use this tool whenever you are unsure."
)

if os.environ.get("AI_MIGRATE_DISABLE_TOOLS", "").lower() in ("true", "1", "yes"):
    LOOKUP_SYMBOL_PROMPT = ""

SYSTEM_MESSAGE = (
    "You are an expert at migrating source code to modern standards. "
    "For each example pair you see, understand the patterns of modernization applied. "
    "Then apply similar modernization patterns to the target code while preserving its functionality. "
    "Provide only the migrated code between triple backticks (```). "
    f"{LOOKUP_SYMBOL_PROMPT}"
)


TOOL_DEFINITIONS = [
    {
        "type": "function",
        "function": {
            "name": FN_LOOKUP_SYMBOL,
            "description": "Get the definition of a symbol from the source code needed to complete the migration",
            "parameters": {
                "type": "object",
                "properties": {
                    "symbol": {
                        "type": "string",
                        "description": "The symbol name to look up (function, constant, etc)",
                    },
                    "import": {
                        "type": "string",
                        "description": "The import statement to add to the source code to get to this symbol",
                    },
                },
                "required": ["symbol", "import"],
            },
        },
    }
]
ROLE_ASSISTANT = "assistant"
ROLE_USER = "user"

LOG_STREAM = contextvars.ContextVar("LOG_STREAM", default=sys.stdout)


def log(*args, **kwargs):
    print(*args, **kwargs, file=LOG_STREAM.get(), flush=True)


@dataclass
class FileContent:
    name: str
    content: str


@dataclass
class MigrationExample:
    name: Optional[str]
    old_files: list[FileContent]
    new_files: list[FileContent]


def read_file_pairs_from(examples_dir: str | Path) -> Iterable[MigrationExample]:
    """
    Read pairs of files from the examples directory, matching .old and .new directories.
    Returns an Iterable of MigrationExample containing all files in each directory pair.

    Sample dir:
    examples/
      example1.old/
        src/
          file1.py
          file2.py
      example1.new/
        src/
          file1.py
          file2.py
      single.old.py
      single.new.py

    Args:
        examples_dir: Directory containing example migration pairs

    Returns:
        Iterable[MigrationExample]: Iterator of migration examples
    """
    examples_dir = Path(examples_dir)

    def get_files_recursively(directory: Path) -> list[FileContent]:
        """Recursively get all files in a directory, maintaining relative paths."""
        files = []
        for item in directory.rglob("*"):
            if item.is_file():
                relative_path = item.relative_to(directory)
                files.append(
                    FileContent(name=str(relative_path), content=item.read_text())
                )
        return files

    old_dirs = [
        d for d in examples_dir.iterdir() if d.is_dir() and d.name.endswith(".old")
    ]

    for old_dir in old_dirs:
        base_name = old_dir.name.removesuffix(".old")
        new_dir = old_dir.parent / f"{base_name}.new"

        if new_dir.exists() and new_dir.is_dir():
            old_files = get_files_recursively(old_dir)
            new_files = get_files_recursively(new_dir)

            yield MigrationExample(
                name=base_name, old_files=old_files, new_files=new_files
            )

    for file in examples_dir.iterdir():
        if file.is_file() and ".old." in file.name:
            base, ext = file.name.split(".old.", 1)
            new_file = file.parent / f"{base}.new.{ext}"

            if new_file.exists():
                old_content = FileContent(
                    name=f"{base}.{ext}", content=file.read_text()
                )
                new_content = FileContent(
                    name=f"{base}.{ext}", content=new_file.read_text()
                )

                yield MigrationExample(
                    name=base, old_files=[old_content], new_files=[new_content]
                )


NBSP_REPLACEMENT_TOKEN = "<LLM-NBSP>"


def escape_nbsp(text: str) -> str:
    """LLMs tokenize these as regular spaces. In source code they need to be preserved."""
    return text.replace("\u00a0", "<LLM-NBSP>")


def un_escape_nbsp(text: str) -> str:
    """LLMs tokenize these as regular spaces. In source code they need to be preserved."""
    return text.replace("<LLM-NBSP>", "\u00a0")


LANGUAGE_EXTENSIONS = {
    "bash": [".sh", ".bash"],
    "python": [".py", ".pyx", ".pyi"],
    "ruby": [".rb", ".rake"],
    "javascript": [".js", ".jsx", ".mjs"],
    "css": [".css", ".cssx"],
    "html": [".html", ".htm", ".htmx"],
    "typescript": [".ts", ".tsx"],
    "java": [".java"],
    "go": [".go"],
    "rust": [".rs"],
    "php": [".php"],
    "swift": [".swift"],
    "kotlin": [".kt", ".kts"],
    "gradle": [".grad"],
    "terraform": [".tf"],
}


def detect_language(filename: str) -> str:
    extension = Path(filename).suffix
    for language, extensions in LANGUAGE_EXTENSIONS.items():
        for e in extensions:
            if e == extension:
                return language
    return ""


def wrap_in_code_block(text: str, filename: str | None) -> str:
    language = detect_language(filename)
    code_block = f"```{language}\n{text.strip()}\n```"
    if filename:
        code_block = f"### `{filename}`\n{code_block}"
    return code_block


def migrate_prompt(example: MigrationExample) -> list[dict]:
    """
    Create a prompt for migrating code based on a MigrationExample.

    Args:
        example: MigrationExample containing old and new files

    Returns:
        List of messages for the LLM conversation
    """
    old_file_contents = "\n\n".join(
        wrap_in_code_block(fc.content, fc.name) for fc in example.old_files
    )
    user_message = {
        "role": ROLE_USER,
        "content": f"Migrate this code to the new format:\n\n{escape_nbsp(old_file_contents)}. "
        "Return the full content for all files mentioned, don't leave anything out. "
        "You can rename a file if necessary.",
    }

    if example.new_files:
        new_file_blocks = [
            wrap_in_code_block(fc.content, fc.name) for fc in example.new_files
        ]
        new_file_contents = "\n\n".join(new_file_blocks)
        assistant_message = {
            "role": ROLE_ASSISTANT,
            "content": f"Here's the migrated code:\n{escape_nbsp(new_file_contents)}",
        }
        return [user_message, assistant_message]
    return [user_message]


def handle_tool_calls(tool_calls: list[dict[str, Any]]) -> list[dict[str, str]]:
    tool_results = []
    for tool_call in tool_calls:
        function = tool_call["function"]
        if function["name"] == FN_LOOKUP_SYMBOL:
            try:
                args = json.loads(function["arguments"])
                symbol = args["symbol"]
                package = args["import"]
                definition = (
                    get_symbol_definition(symbol, package)
                    or "Symbol not found. It might not be available in this repository."
                )
                log(f"Looking up symbol: {symbol}")
                log(">>", definition)

                tool_results.append(
                    {
                        "tool_call_id": tool_call["id"],
                        "output": definition,
                    }
                )
            except Exception as e:
                tool_results.append(
                    {
                        "tool_call_id": tool_call["id"],
                        "output": f"Error processing tool call: {str(e)}",
                    }
                )
        else:
            tool_results.append(
                {
                    "tool_call_id": tool_call["id"],
                    "output": f"Unknown tool call: {tool_call['function']['name']}",
                }
            )
    return tool_results


def messages_and_tools(
    examples: list[MigrationExample],
    target: MigrationExample,
    system_prompt: str,
) -> tuple[list, list]:
    messages = [{"role": "system", "content": system_prompt}]

    for example in examples:
        messages.extend(migrate_prompt(example))

    messages.extend(migrate_prompt(target))

    # Check if tool calling should be disabled
    if os.environ.get("AI_MIGRATE_DISABLE_TOOLS", "").lower() in ("true", "1", "yes"):
        return messages, []
    else:
        return messages, TOOL_DEFINITIONS


async def call_llm(
    client: DefaultClient, messages: list, tools: list, temperature=0.1
) -> tuple[dict, list[dict]]:
    """Call LLM for completions

    Returns the response and the messages so far
    """
    while True:
        response, messages = await client.generate_completion(
            messages=messages, tools=tools, temperature=temperature
        )

        assistant_message = response["choices"][0]["message"]

        if not (tool_calls := assistant_message.get("tool_calls")):
            return response, messages

        tool_results = handle_tool_calls(tool_calls)

        messages.append(assistant_message)
        for result in tool_results:
            messages.append(
                {
                    "role": "tool",
                    "tool_call_id": result["tool_call_id"],
                    "content": result["output"],
                }
            )

        for message in messages:
            if message.get("content") == "":
                del message["content"]


def filter_lines(text, match):
    return "\n".join(line for line in text.splitlines() if match in line)


async def subprocess_run(cmd, prefix=None, **kwargs) -> str:
    if not kwargs.get("stderr"):
        kwargs["stderr"] = subprocess.STDOUT
        kwargs["stdout"] = subprocess.PIPE
    check = False
    if "check" in kwargs:
        check = kwargs.pop("check")
    env = kwargs.pop("env", {})
    env = {**os.environ, "PYTHONUNBUFFERED": "1", **env}
    process = await asyncio.create_subprocess_exec(
        *cmd,
        **kwargs,
        env=env,
    )
    stdout = []
    if not prefix:
        prefix = cmd[0]
    while not process.stdout.at_eof():
        data = await process.stdout.readline()
        if not data:
            break
        line = data.decode().removesuffix("\n")
        log(f"[{prefix}] {line}")
        stdout.append(line)
    await process.wait()
    if check and process.returncode != 0:
        raise subprocess.CalledProcessError(process.returncode, cmd)
    return "\n".join(stdout)


async def run(
    target_files: list[str],
    system_prompt,
    examples_dir,
    verify_cmd,
    pre_verify_cmd,
    log_stream,
    local_worktrees,
    llm_fakes,
    target_dir: str = "",
    target_basename: str = "",
    dont_create_evals: bool = False,
):
    """Run the migration process on the target files.
    Args:
        target_files: List of files to migrate
        system_prompt: System prompt for the LLM
        examples_dir: Directory containing example migrations
        verify_cmd: Command to verify the migrated files
        pre_verify_cmd: Command to run before migration
        log_stream: Stream to write logs to
        local_worktrees: Whether to create worktrees locally
        dont_create_evals: If True, don't automatically create evaluations after successful migrations
    """
    if log_stream:
        LOG_STREAM.set(log_stream)

    source_git_root = Path(
        (
            await subprocess_run(["git", "rev-parse", "--show-toplevel"], check=True)
        ).strip()
    )
    source_start_point = (
        await subprocess_run(["git", "rev-parse", "HEAD"], check=True)
    ).strip()

    # If target_dir is specified, use it as the git root for the output
    if target_dir:
        target_dir_path = Path(target_dir)
        try:
            if not target_dir_path.exists():
                target_dir_path.mkdir(parents=True, exist_ok=True)

            target_git_root = Path(
                (
                    await subprocess_run(
                        ["git", "rev-parse", "--show-toplevel"],
                        check=True,
                        cwd=str(target_dir_path),
                    )
                ).strip()
            )
            start_point = (
                await subprocess_run(
                    ["git", "rev-parse", "HEAD"], check=True, cwd=str(target_git_root)
                )
            ).strip()

            target_dir_rel_path = target_dir_path.resolve().relative_to(
                target_git_root.resolve()
            )

            git_root = target_git_root
        except Exception as e:
            log(f"Error finding git repository: {e}")
            raise ValueError(
                f"Error: target_dir '{target_dir}' is not in a git repository."
            )
    else:
        # Use the source repository as the git root
        git_root = source_git_root
        start_point = source_start_point
        target_dir_rel_path = None

    # Get the relative paths of the target files
    targets_in_repo = [
        Path(target_file).resolve().relative_to(source_git_root)
        for target_file in target_files
    ]

    file_flat = FileGroup(files=target_files).group_name()
    worktree_name = f"ai-migrator-worktree-{git_root.name}-{file_flat}"
    if local_worktrees:
        worktree_root = git_root.parent / "ai-migrator-worktrees" / worktree_name
    else:
        worktree_root = Path(tempfile.gettempdir()) / worktree_name
    branch = f"ai-migrator/{file_flat}"

    if not worktree_root.exists():
        await subprocess_run(
            ["git", "worktree", "add", worktree_root, "HEAD"],
            check=True,
            cwd=git_root,
        )
    await subprocess_run(
        ["git", "checkout", "--force", "-B", branch, start_point],
        check=True,
        cwd=worktree_root,
    )

    # If using target_dir, read files from original location instead of worktree
    target_root = source_git_root if target_dir else worktree_root
    targets_files = [target_root / target_in_repo for target_in_repo in targets_in_repo]

    return await _run(
        targets_files,
        system_prompt,
        examples_dir,
        verify_cmd,
        pre_verify_cmd,
        worktree_root,
        llm_fakes,
        dont_create_evals,
        target_dir=target_dir,
        target_dir_rel_path=target_dir_rel_path,
        target_basename=target_basename,
    )


@dataclass
class CodeBlock:
    filename: str | None
    code: str


@dataclass
class CodeResponseResult:
    code_blocks: list[CodeBlock]
    other_text: str


def extract_code_blocks(markdown, replacement="<code>") -> CodeResponseResult:
    lines = markdown.splitlines()
    filename = None
    line_it = iter(lines)
    result = CodeResponseResult([], "")
    other_text = []

    for line in line_it:
        if line.lstrip().startswith("### ") and line.count("`") == 2:
            start = line.find("`")
            end = line.find("`", start + 1)
            filename = line[start + 1 : end]
        elif line.lstrip().startswith("```"):
            code = []
            for line in line_it:
                if line.lstrip().startswith("```"):
                    break
                code.append(line)
            result.code_blocks.append(CodeBlock(filename, "\n".join(code)))
            filename = None
            other_text.append(replacement)
        else:
            other_text.append(line)

    if other_text:
        result.other_text = "\n".join(other_text)

    return result


class FailedPreVerification(Exception):
    pass


async def remove_worktree(worktree_root: str | Path):
    await asyncio.to_thread(shutil.rmtree, worktree_root, ignore_errors=True)


def build_messages(
    initial_messages: list[dict], iteration_messages: list[list[dict]]
) -> list[dict]:
    messages = initial_messages
    for iteration_message in iteration_messages:
        messages.extend(iteration_message)
    return messages


async def _run(
    target_files,
    system_prompt,
    examples_dir,
    verify_cmd,
    pre_verify_cmd,
    worktree_root,
    llm_fakes,
    dont_create_evals=False,
    target_dir=None,
    target_dir_rel_path=None,
    target_basename=None,
):
    if llm_fakes:
        client = FakeLLMClient(llm_fakes)
    else:
        client = DefaultClient()

    verify_cmd = verify_cmd.split()

    examples = [*read_file_pairs_from(examples_dir)]
    if not examples:
        raise FileNotFoundError("No valid example pairs found in examples directory")

    system_prompt = Path(system_prompt).read_text()

    # TODO: Have some kind of configuration driven controls for how basename is transformed
    if target_basename:
        target_basename = (
            target_basename.replace("-", " ").replace("_", " ").title().replace(" ", "")
        )

    # Create target MigrationExample
    target_file_contents = []
    for i, target_file in enumerate(target_files):
        full_path = Path(target_file).absolute()

        if target_dir:
            read_path = Path(target_files[i]).absolute()
            short_name = full_path.name
            content = read_path.read_text()
        else:
            short_name = full_path.relative_to(worktree_root)
            content = full_path.read_text()

        target_file_contents.append(FileContent(name=str(short_name), content=content))

    target = MigrationExample(name=None, old_files=target_file_contents, new_files=[])

    messages, tools = messages_and_tools(examples, target, system_prompt)
    all_files_to_verify = set()

    iteration_messages = []

    if pre_verify_cmd:
        log("Running pre-verification")
        try:
            await subprocess_run(
                [*pre_verify_cmd.split(), *target_files],
                prefix="pre-verify",
                check=True,
                cwd=worktree_root,
            )
        except subprocess.CalledProcessError:
            log(
                "Pre-verification failed. The migration cannot continue until all files pass the pre-verify step"
            )
            raise FailedPreVerification(f"file {target_files} failed pre-verification")

    if not target_dir:
        for target_file in target_files:
            os.remove(target_file)
            await subprocess_run(
                ["git", "add", target_file],
                cwd=worktree_root,
            )

    for tries in range(int(os.getenv("AI_MIGRATE_MAX_TRIES", 10))):
        log(f"[agent] Running migration attempt {tries + 1}")

        messages = build_messages(messages, iteration_messages)
        log(f"[agent] Messages length: {client.count_tokens(messages)}")
        while 0 < client.max_context_tokens() < client.count_tokens(messages):
            log(f"Trimming iteration messages: {len(iteration_messages)}")
            # Fall back to 3 examples + trim iterations 1 at a time.
            messages, _ = messages_and_tools(examples[:3], target, system_prompt)
            iteration_messages = iteration_messages[1:]
            messages = build_messages(messages, iteration_messages)

        response, messages = await call_llm(client, messages, tools)

        response_text = response["choices"][0]["message"]["content"]
        parsed_result = extract_code_blocks(response_text)

        iteration_message = []
        if not parsed_result.code_blocks:
            iteration_message.append({"role": ROLE_ASSISTANT, "content": response_text})
            iteration_message.append(
                {
                    "role": ROLE_USER,
                    "content": "Include the full, complete code block. Do not omit any part of the file",
                }
            )
            continue

        written_files = set()
        for code_block in parsed_result.code_blocks:
            migrated_code = un_escape_nbsp(code_block.code)

            if code_block.filename:
                written_files.add(code_block.filename)
                if target_dir:
                    output_path = (
                        Path(worktree_root)
                        / target_dir_rel_path
                        / target_basename
                        / code_block.filename
                    )
                else:
                    output_path = Path(worktree_root) / code_block.filename
                output_path.parent.mkdir(parents=True, exist_ok=True)

                with open(output_path, "w") as f:
                    f.write(migrated_code)

        all_files_to_verify |= written_files

        # Add the files to verify with the correct paths
        if target_dir:
            full_verify_cmd = [
                *verify_cmd,
                str(Path(target_dir_rel_path) / target_basename),
            ]
        else:
            full_verify_cmd = [
                *verify_cmd,
                *[str(Path(worktree_root) / f) for f in all_files_to_verify],
            ]

        log(f"Running verification: {full_verify_cmd}")
        verify_process = await asyncio.create_subprocess_exec(
            *full_verify_cmd,
            cwd=worktree_root,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = await verify_process.communicate()

        status = "pass" if verify_process.returncode == 0 else "fail"

        commit_message = f"Migration attempt {tries + 1} {status=}:\n\nLLM response:\n{parsed_result.other_text}"

        for file in written_files:
            if target_dir:
                file_path = (
                    Path(worktree_root) / target_dir_rel_path / target_basename / file
                )
                git_path = Path(file_path).relative_to(worktree_root)
                await subprocess_run(
                    ["git", "add", git_path],
                    cwd=worktree_root,
                )
            else:
                await subprocess_run(
                    ["git", "add", file],
                    cwd=worktree_root,
                )

        await subprocess_run(
            ["git", "commit", "--allow-empty", "-m", commit_message],
            check=True,
            cwd=worktree_root,
            env={**os.environ, **environment_variables()},
        )

        verification_output = (stderr or stdout or b"").decode()
        await subprocess_run(
            [
                "git",
                "notes",
                "--ref=migrator-verify",
                "add",
                "-f",
                "-m",
                verification_output,
            ],
            check=True,
            cwd=worktree_root,
        )

        if verify_process.returncode == 0:
            log("Verification successful")

            if not dont_create_evals:
                original_contents = {}
                for target_file in target_files:
                    full_path = Path(target_file).absolute()
                    short_name = full_path.relative_to(worktree_root)
                    try:
                        with open(full_path, "r") as f:
                            original_contents[str(short_name)] = f.read()
                    except Exception as e:
                        log(f"Error reading {full_path}: {e}")

                transformed_contents = {}
                for file_path in written_files:
                    full_path = Path(worktree_root) / file_path
                    try:
                        with open(full_path, "r") as f:
                            transformed_contents[file_path] = f.read()
                    except Exception as e:
                        log(f"Error reading transformed file {full_path}: {e}")

                try:
                    project_dir = Path(examples_dir).parent

                    eval_manifest = Manifest(
                        files=[
                            FileEntry(filename=fname, result="pass")
                            for fname in written_files
                        ],
                        verify_cmd=" ".join(verify_cmd)
                        if isinstance(verify_cmd, list)
                        else verify_cmd,
                    )
                    if pre_verify_cmd:
                        eval_manifest.pre_verify_cmd = pre_verify_cmd

                    eval_dir = generate_eval_from_migration(
                        project_dir,
                        original_contents,
                        transformed_contents,
                        eval_manifest,
                    )
                    log(f"Created evaluation at {eval_dir}")
                except Exception as e:
                    log(f"Error creating evaluation: {e}")
                    log(f"Exception type: {type(e).__name__}")

            await remove_worktree(worktree_root)
            break
        log("Verification failed:")
        for line in verification_output.splitlines():
            log(f"[verify] {line}")

        lookup_symbol_prompt = (
            f"Use the {FN_LOOKUP_SYMBOL} tool to find the definition of any symbols related to this problem. "
            "Then apply the necessary changes to the code. Don't guess or assume functionality."
            "From now on, only re-write files, don't rename them."
        )
        if os.environ.get("AI_MIGRATE_DISABLE_TOOLS", "").lower() in (
            "true",
            "1",
            "yes",
        ):
            lookup_symbol_prompt = ""

        iteration_message.append({"role": ROLE_ASSISTANT, "content": response_text})
        iteration_message.append(
            {
                "role": ROLE_USER,
                "content": f"The code did not compile. The error was: {verification_output}. "
                f"{lookup_symbol_prompt}",
            }
        )
        iteration_messages.append(iteration_message)

    else:
        raise ValueError("Migration failed: Out of tries")



import re
import subprocess
from typing import Literal

from ai_migrate.manifest import Manifest

Status = Literal["pass", "fail", "?"]


def get_branches(
    manifest: Manifest | None = None,
) -> list[tuple[str, str, Status, str]]:
    result = subprocess.run(
        [
            "git",
            "branch",
            "--format=%(objectname:short) %(refname:short) %(contents:lines=1)",
        ],
        capture_output=True,
        text=True,
        check=True,
    )
    branches = result.stdout.strip().split("\n")

    in_manifest = {f.group_name() for f in manifest.files} if manifest else {}
    matching = []
    for line in branches:
        sha, branch, msg = line.split(maxsplit=2)
        if branch.startswith("ai-migrator/"):
            if not in_manifest or branch.removeprefix("ai-migrator/") in in_manifest:
                pattern = r"Migration attempt \d+ status='(.*)':"
                match = re.search(pattern, msg)
                status = match.group(1) if match else "?"
                matching.append((sha, branch, status, msg))

    return matching


def get_worktrees() -> list[tuple[str, str, bool]]:
    result = subprocess.run(
        ["git", "worktree", "list"],
        capture_output=True,
        text=True,
        check=True,
    )
    worktrees = []
    for row in result.stdout.strip().splitlines():
        worktree, sha, branch_with_brackets, *rest = row.split()
        assert branch_with_brackets[0] == "[" and branch_with_brackets[-1] == "]"
        branch = branch_with_brackets[1:-1]
        worktrees.append(
            (
                worktree,
                branch,
                "prunable" in rest,
            )
        )
    return worktrees



import asyncio
import json
import re
import subprocess
from pathlib import Path
from typing import Dict, List, Tuple

from ai_migrate.llm_providers import DefaultClient


async def _run_gh_command(args: list) -> str:
    """Run a GitHub CLI command and return its output.

    Args:
        args: List of command arguments to pass to gh

    Returns:
        The command output as a string

    Raises:
        subprocess.CalledProcessError: If the command fails
    """
    process = await asyncio.create_subprocess_exec(
        "gh",
        *args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = await process.communicate()

    if process.returncode != 0:
        raise subprocess.CalledProcessError(
            process.returncode, ["gh"] + args, output=stdout, stderr=stderr
        )

    return stdout.decode()


async def get_pr_details(pr_url: str) -> Dict:
    """Get details about a PR using the GitHub CLI.

    Args:
        pr_url: The PR URL or reference (e.g., 'https://github.com/owner/repo/pull/123' or 'owner/repo#123')

    Returns:
        A dictionary with PR details
    """
    json_output = await _run_gh_command(
        [
            "pr",
            "view",
            pr_url,
            "--json",
            "title,body,files,commits,additions,deletions,changedFiles",
        ]
    )

    return json.loads(json_output)


async def get_file_diff(pr_url: str, file_path: str) -> str:
    """Get the diff for a specific file in a PR.

    Args:
        pr_url: The PR URL or reference
        file_path: The path to the file

    Returns:
        The diff content as a string
    """

    file_list = await _run_gh_command(["pr", "diff", pr_url, "--name-only"])

    files = file_list.strip().split("\n")
    if file_path not in files:
        raise ValueError(f"File {file_path} not found in PR {pr_url}")

    return await _run_gh_command(["pr", "diff", pr_url, file_path])


async def get_file_content(pr_url: str, file_path: str, base: bool = False) -> str:
    """Get the content of a file from a PR.

    Args:
        pr_url: The PR URL or reference
        file_path: The path to the file
        base: If True, get the base (before) version, otherwise get the head (after) version

    Returns:
        The file content as a string
    """
    # Approach 1: Use gh pr diff to get the file content
    try:
        diff_content = await _run_gh_command(["pr", "diff", pr_url, "--patch"])

        file_diffs = diff_content.split("diff --git ")
        for file_diff in file_diffs:
            if f"a/{file_path}" in file_diff or f"b/{file_path}" in file_diff:
                # Found the right file
                lines = file_diff.splitlines()
                content_lines = []
                in_content = False

                for line in lines:
                    if line.startswith("+++") and base:
                        in_content = True
                        continue
                    elif line.startswith("---") and not base:
                        in_content = True
                        continue

                    if in_content:
                        if line.startswith("+") and not base:
                            content_lines.append(line[1:])
                        elif line.startswith("-") and base:
                            content_lines.append(line[1:])
                        elif not line.startswith("+") and not line.startswith("-"):
                            content_lines.append(line)

                if content_lines:
                    return "\n".join(content_lines)
    except Exception as e:
        print(f"Approach 1 failed: {e}")

    # Approach 2: Use GitHub API to get file content
    try:
        branch_json = await _run_gh_command(
            ["pr", "view", pr_url, "--json", "baseRefName,headRefName"]
        )

        branch_data = json.loads(branch_json)
        ref = branch_data["baseRefName"] if base else branch_data["headRefName"]

        content_json = await _run_gh_command(
            ["api", f"repos/:owner/:repo/contents/{file_path}?ref={ref}"]
        )

        content_data = json.loads(content_json)
        import base64

        return base64.b64decode(content_data["content"]).decode("utf-8")
    except Exception as e:
        print(f"Approach 2 failed: {e}")

    raise ValueError(f"Unable to retrieve content for {file_path}")


async def generate_system_prompt(pr_details: Dict, description: str) -> str:
    """Generate a system prompt based on PR details and a description.

    Args:
        pr_details: The PR details
        description: A short description of the migration task

    Returns:
        The generated system prompt
    """
    client = DefaultClient()

    system_prompt = """You are an expert at creating system prompts for code migration tasks.
Given information about a pull request that demonstrates a code migration pattern, 
create a clear, concise system prompt that will guide an AI to perform similar migrations on other files."""

    user_prompt = f"""
I need to create a system prompt for a code migration task. Here are the details:

Description: {description}

Pull Request Title: {pr_details["title"]}
Pull Request Description: {pr_details["body"]}

Files Changed: {len(pr_details["files"])}
Total Additions: {pr_details["additions"]}
Total Deletions: {pr_details["deletions"]}

Create a system prompt that clearly explains:
1. The migration goal
2. Key patterns to look for
3. How to transform the code
4. Any important constraints or considerations

The system prompt should be concise but comprehensive, focusing on the patterns demonstrated in the PR.
"""

    return await client.generate_text(system_prompt, user_prompt)


async def extract_example_patterns(
    pr_url: str, pr_details: Dict
) -> List[Tuple[str, str]]:
    """Extract example patterns from a PR.

    Returns:
        A list of (before, after) example pairs
    """
    client = DefaultClient()

    files_to_analyze = pr_details.get("files", [])
    if not files_to_analyze and "changedFiles" in pr_details:
        try:
            files_json = await _run_gh_command(
                ["pr", "view", pr_url, "--json", "files"]
            )
            files_data = json.loads(files_json)
            files_to_analyze = files_data.get("files", [])
        except Exception:
            pass

    if not files_to_analyze:
        try:
            file_paths = await _run_gh_command(["pr", "diff", pr_url, "--name-only"])
            files_to_analyze = [
                {"path": path} for path in file_paths.strip().split("\n")
            ]
        except Exception as e:
            print(f"Error getting files from diff: {e}")

    files_to_analyze = files_to_analyze[:5]

    examples = []
    for file_info in files_to_analyze:
        if not isinstance(file_info, dict) or "path" not in file_info:
            continue

        file_path = file_info["path"]

        if "status" in file_info and (
            file_info["status"] == "added" or file_info["status"] == "deleted"
        ):
            continue

        try:
            before_content = await get_file_content(pr_url, file_path, base=True)
            after_content = await get_file_content(pr_url, file_path, base=False)

            if "not available" in before_content and "not available" in after_content:
                print(f"Skipping {file_path} - content not available")
                continue

            system_prompt = """You are an expert at identifying code migration patterns.
Given a before and after version of a file, extract the minimal, representative example that 
clearly demonstrates the migration pattern. Focus on the core transformation pattern, not incidental changes."""

            user_prompt = f"""
Analyze these before and after versions of a file and extract a minimal example that demonstrates the key migration pattern:

BEFORE:
```
{before_content}
```

AFTER:
```
{after_content}
```

Extract just the essential parts that show the migration pattern. The example should be as small as possible while still 
demonstrating the pattern clearly. Remove any code that doesn't directly demonstrate the migration pattern.

Respond with two code blocks labeled BEFORE and AFTER containing your minimal examples.
"""

            response = await client.generate_text(system_prompt, user_prompt)

            before_match = re.search(r"BEFORE:\s*```.*?\n(.*?)```", response, re.DOTALL)
            after_match = re.search(r"AFTER:\s*```.*?\n(.*?)```", response, re.DOTALL)

            if before_match and after_match:
                before_example = before_match.group(1).strip()
                after_example = after_match.group(1).strip()
                examples.append((before_example, after_example))
        except Exception as e:
            print(f"Error processing file {file_path}: {e}")

    if not examples and pr_details.get("title") and pr_details.get("body"):
        try:
            system_prompt = """You are an expert at creating code examples for migration patterns.
Given a PR title and description, create a minimal example that demonstrates the migration pattern described."""

            user_prompt = f"""
Create a minimal example that demonstrates the migration pattern described in this PR:

Title: {pr_details["title"]}
Description: {pr_details["body"]}

Create two code blocks labeled BEFORE and AFTER that show the migration pattern.
The examples should be minimal but clearly demonstrate the key changes involved in this migration.
"""

            response = await client.generate_text(system_prompt, user_prompt)

            before_match = re.search(r"BEFORE:\s*```.*?\n(.*?)```", response, re.DOTALL)
            after_match = re.search(r"AFTER:\s*```.*?\n(.*?)```", response, re.DOTALL)

            if before_match and after_match:
                before_example = before_match.group(1).strip()
                after_example = after_match.group(1).strip()
                examples.append((before_example, after_example))
        except Exception as e:
            print(f"Error creating example from PR description: {e}")

    return examples


async def save_examples(
    examples: List[Tuple[str, str]], examples_dir: Path, file_extension: str = "java"
):
    """Save example pairs to files.

    Args:
        examples: List of (before, after) example pairs
        examples_dir: Directory to save examples to
        file_extension: File extension to use for examples
    """
    examples_dir.mkdir(parents=True, exist_ok=True)

    for i, (before, after) in enumerate(examples):
        base_name = f"Example{i + 1}"
        before_file = examples_dir / f"{base_name}.old.{file_extension}"
        after_file = examples_dir / f"{base_name}.new.{file_extension}"

        before_file.write_text(before)
        after_file.write_text(after)

    print(f"Saved {len(examples)} example pairs to {examples_dir}")


async def generate_verify_script(
    pr_details: Dict, project_dir: Path, file_extension: str
) -> str:
    """Generate a verification script based on PR details.

    Args:
        pr_details: The PR details
        project_dir: The project directory
        file_extension: The file extension of the migrated files

    Returns:
        The generated verification script
    """
    client = DefaultClient()

    system_prompt = """You are an expert at creating verification scripts for code migration tasks.
Given information about a pull request that demonstrates a code migration pattern, create a Python script
that can verify if a file has been properly migrated according to the pattern."""

    user_prompt = f"""
I need to create a verification script for a code migration task. Here are the details:

Pull Request Title: {pr_details["title"]}
Pull Request Description: {pr_details["body"]}

Files Changed: {len(pr_details["files"])}
File Extension: {file_extension}

Create a Python script that:
1. Takes a file path as input
2. Verifies if the file has been properly migrated according to the pattern
3. Returns a non-zero exit code if verification fails

The script should be saved as verify.py in the project directory and should be runnable as:
python verify.py <file_path>

For pre-verification (to gather information before migration), it should support:
python verify.py --pre <file_path>

Focus on creating a practical verification script that checks for the key aspects of the migration.
"""

    verify_script = await client.generate_text(system_prompt, user_prompt)

    code_match = re.search(r"```python\n(.*?)```", verify_script, re.DOTALL)
    if code_match:
        verify_script = code_match.group(1)

    return verify_script


async def setup_project_from_pr(
    pr_url: str, project_path: str, description: str, file_extension: str = "java"
) -> None:
    """Set up a migration project based on a PR.

    Args:
        pr_url: The PR URL or reference (e.g., 'https://github.com/owner/repo/pull/123' or 'owner/repo#123')
        project_path: Path where the project should be created
        description: Short description of the migration task
        file_extension: File extension for the migrated files
    """
    project_dir = Path(project_path).expanduser()

    project_dir.mkdir(parents=True, exist_ok=True)
    examples_dir = project_dir / "examples"
    examples_dir.mkdir(exist_ok=True)
    evals_dir = project_dir / "evals"
    evals_dir.mkdir(exist_ok=True)

    print(f"Created project directory structure at {project_dir}")

    try:
        print("Fetching PR details...")
        pr_details = await get_pr_details(pr_url)

        print("Generating system prompt...")
        system_prompt = await generate_system_prompt(pr_details, description)
        system_prompt_file = project_dir / "system_prompt.md"
        system_prompt_file.write_text(system_prompt)
        print(f"Saved system prompt to {system_prompt_file}")

        print("Extracting example patterns...")
        examples = await extract_example_patterns(pr_url, pr_details)
        if not examples:
            print(
                "Warning: No example patterns were extracted. Creating a simple example..."
            )
            examples = [
                (
                    "// Old version\npublic class Example {\n    // TODO: Add your code here\n}",
                    "// New version\npublic class Example {\n    // TODO: Migrated code here\n}",
                )
            ]
        await save_examples(examples, examples_dir, file_extension)

        print("Generating verification script...")
        verify_script = await generate_verify_script(
            pr_details, project_dir, file_extension
        )
        verify_script_file = project_dir / "verify.py"
        verify_script_file.write_text(verify_script)
        print(f"Saved verification script to {verify_script_file}")

        print(f"Project setup complete at {project_dir}")
    except Exception as e:
        print(f"Error during project setup: {e}")
        from .migrate import SYSTEM_MESSAGE
        from .manifest import SYSTEM_PROMPT_FILE, VERIFY_SCRIPT_FILE

        system_prompt_file = project_dir / SYSTEM_PROMPT_FILE
        if not system_prompt_file.exists():
            system_prompt = f"{SYSTEM_MESSAGE}\n\n# Migration Task\n\n{description}"
            system_prompt_file.write_text(system_prompt)
            print(f"Created minimal system prompt at {system_prompt_file}")

        verify_script_file = project_dir / VERIFY_SCRIPT_FILE
        if not verify_script_file.exists():
            verify_script = (
                "import sys\n\n"
                "def main():\n"
                "    # TODO: Implement verification\n"
                '    print(f"Verifying {sys.argv[1]}...")\n'
                "    return 0\n\n"
                'if __name__ == "__main__":\n'
                "    sys.exit(main())\n"
            )
            verify_script_file.write_text(verify_script)
            print(f"Created minimal verification script at {verify_script_file}")

        print(
            "Project setup completed with errors. Some files may need manual editing."
        )



import os
import time
import json
import logging
import asyncio
import subprocess
import tempfile
import shutil
from pathlib import Path
from typing import Optional

from pydantic import BaseModel

from .manifest import Manifest, FileEntry

logger = logging.getLogger("ai_migrate.eval_generator")
GITHUB_DOMAIN = "https://github.com/"


def generate_eval_from_migration(
    project_template,
    source_files,
    transformed_files,
    manifest: Manifest,
    eval_name=None,
):
    """
    Generate a new evaluation test case from a successful migration.

    Args:
        project_template: Path to the project template directory
        source_files: Dict mapping filenames to original content
        transformed_files: Dict mapping filenames to transformed content
        manifest: The manifest used for the migration (must be a Manifest object)
        eval_name: Optional name for the evaluation, defaults to timestamp-based name

    Returns:
        Path to the created evaluation directory
    """
    if eval_name is None:
        eval_name = f"auto_generated_{int(time.time())}"

    project_dir = Path(project_template)
    eval_dir = project_dir / "evals" / eval_name

    os.makedirs(eval_dir / "source", exist_ok=True)

    for filename, content in source_files.items():
        rel_path = Path(filename)
        file_path = eval_dir / "source" / rel_path.name
        file_path.parent.mkdir(parents=True, exist_ok=True)
        with open(file_path, "w") as f:
            f.write(content)

    with open(eval_dir / "manifest.json", "w") as f:
        f.write(manifest.model_dump_json(indent=2))

    logger.info(f"Generated new evaluation: {eval_dir}")
    return eval_dir


class RepoInfo(BaseModel):
    """Pydantic model for GitHub repository information extracted from URLs."""

    owner: Optional[str] = None
    name: Optional[str] = None
    pr_number: Optional[str] = None

    @property
    def full_repo(self) -> Optional[str]:
        """Return the full repository name as owner/name if both are available."""
        if self.owner and self.name:
            return f"{self.owner}/{self.name}"
        return None


def extract_repo_info_from_url(url: str) -> RepoInfo:
    """Extract repository owner and name from various URL formats.

    Args:
        url: A GitHub URL or repo reference (e.g., 'https://github.com/owner/repo/pull/123' or 'owner/repo#123')

    Returns:
        RepoInfo: A Pydantic model containing extracted information
    """
    info = RepoInfo()

    if "#" in url:
        info.pr_number = url.split("#")[-1]
    elif "/pull/" in url:
        info.pr_number = url.split("/pull/")[-1]

    if info.pr_number and not info.pr_number.isdigit():
        logger.warning(
            f"PR number '{info.pr_number}' contains non-numeric characters. Using as is."
        )

    if "/" in url:
        parts = url.split("/")
        if len(parts) >= 5 and parts[2] == "github.com":
            info.owner = parts[3]
            info.name = parts[4].split("#")[0].split("/")[0]
        elif len(parts) >= 2:
            info.owner = parts[0]
            info.name = parts[1].split("#")[0]

    return info


async def generate_eval_from_pr(pr_url: str, project_template: str) -> Path:
    """Generate evaluation files from a PR.

    Args:
        pr_url: The PR URL or reference (e.g., 'https://github.com/owner/repo/pull/123' or 'owner/repo#123')
        project_template: Path to the project template directory

    Returns:
        Path to the created evaluation directory
    """
    project_dir = Path(project_template).expanduser()
    evals_dir = project_dir / "evals"
    evals_dir.mkdir(exist_ok=True)

    original_pr_url = pr_url

    repo_info = extract_repo_info_from_url(original_pr_url)

    pr_number = repo_info.pr_number or str(int(time.time()))

    try:
        logger.info(f"Fetching PR details for: {original_pr_url}")
        result = await asyncio.create_subprocess_exec(
            "gh",
            "pr",
            "view",
            original_pr_url,
            "--json",
            "title,headRepository,baseRefName,headRefName,baseRefOid,headRefOid,files",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = await result.communicate()

        if result.returncode != 0:
            error_msg = stderr.decode() if stderr else "Unknown error"
            logger.error(f"Failed to get PR details: {error_msg}")

            if "not authenticated" in error_msg or "Unauthorized" in error_msg:
                raise ValueError(
                    f"GitHub authentication error. Please run 'gh auth login' to authenticate with GitHub.\n"
                    f"Error details: {error_msg}"
                )

            if (
                "Could not resolve to a Repository" in error_msg
                or "Could not resolve to a PullRequest" in error_msg
            ):
                msg = "Could not access the repository"
                if repo_info.full_repo:
                    msg += f" {repo_info.full_repo}"
                elif repo_info.name:
                    msg += f" {repo_info.name}"
                msg += (
                    ". Make sure you have access to the repository and the PR exists.\n"
                )
                msg += f"Error details: {error_msg}"
                raise ValueError(msg)

            raise ValueError(f"Failed to get PR details: {error_msg}")

        pr_details = json.loads(stdout.decode())
    except json.JSONDecodeError:
        logger.error("Invalid JSON response from GitHub CLI")
        raise ValueError(
            "Invalid response from GitHub CLI. Please check your GitHub CLI installation and authentication."
        )

    repo_name = f"pr-{pr_number}"
    if "headRepository" in pr_details and "name" in pr_details["headRepository"]:
        repo_name = pr_details["headRepository"]["name"]

    eval_name = f"{repo_name}-{int(time.time())}"
    eval_dir = evals_dir / eval_name
    eval_dir.mkdir(exist_ok=True)

    base_ref = pr_details.get("baseRefOid", "")

    changed_files = []
    if "files" in pr_details and isinstance(pr_details["files"], list):
        changed_files = [
            file_info["path"]
            for file_info in pr_details["files"]
            if isinstance(file_info, dict) and "path" in file_info
        ]

    repo_url = ""
    if "headRepository" in pr_details and "url" in pr_details["headRepository"]:
        repo_url = pr_details["headRepository"]["url"]

        if repo_url.startswith(GITHUB_DOMAIN):
            repo_url = f"github@github.com:{repo_url[len(GITHUB_DOMAIN) :]}.git"

    manifest = Manifest(
        eval_target_repo_remote=repo_url,
        eval_target_repo_ref=base_ref,
        files=[
            FileEntry(filename=file_path, result="?")
            for file_path in changed_files[:10]
        ],
    )

    manifest_file = eval_dir / "manifest.json"
    with open(manifest_file, "w") as f:
        f.write(manifest.model_dump_json(indent=2))

    source_dir = eval_dir / "source"
    source_dir.mkdir(exist_ok=True)

    if base_ref and changed_files:
        with tempfile.TemporaryDirectory() as temp_dir:
            clone_url = (
                repo_url
                if repo_url
                else f"https://github.com/{pr_details['headRepository']['owner']['login']}/{pr_details['headRepository']['name']}.git"
            )

            clone_result = await asyncio.create_subprocess_exec(
                "git",
                "clone",
                "--depth=1",
                clone_url,
                temp_dir,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            clone_stdout, clone_stderr = await clone_result.communicate()

            if clone_result.returncode != 0:
                error_msg = clone_stderr.decode() if clone_stderr else "Unknown error"
                logger.warning(f"Failed to clone repository: {error_msg}")

                raise subprocess.CalledProcessError(
                    clone_result.returncode,
                    ["git", "clone", "--depth=1", clone_url, temp_dir],
                    output=clone_stdout,
                    stderr=clone_stderr,
                )

            checkout_result = await asyncio.create_subprocess_exec(
                "git",
                "checkout",
                base_ref,
                cwd=temp_dir,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            checkout_stdout, checkout_stderr = await checkout_result.communicate()

            if checkout_result.returncode != 0:
                error_msg = (
                    checkout_stderr.decode() if checkout_stderr else "Unknown error"
                )
                logger.warning(f"Failed to checkout base commit: {error_msg}")

                raise subprocess.CalledProcessError(
                    checkout_result.returncode,
                    ["git", "checkout", base_ref],
                    output=checkout_stdout,
                    stderr=checkout_stderr,
                )

            for file_path in changed_files:
                src_file = Path(temp_dir) / file_path
                if src_file.exists():
                    dst_file = source_dir / file_path
                    dst_file.parent.mkdir(parents=True, exist_ok=True)
                    try:
                        shutil.copy2(src_file, dst_file)
                    except (IOError, OSError) as e:
                        logger.warning(f"Failed to copy file {file_path}: {e}")

    logger.info(f"Generated evaluation at {eval_dir}")
    return eval_dir



from itertools import cycle
from pathlib import Path
from typing import Any, Dict, List, Union


class FakeLLMClient:
    def __init__(self, directory):
        self.responses = cycle([f.read_text() for f in Path(directory).iterdir()])

    async def generate_completion(
        self, messages, tools=None, temperature=0.0, max_tokens=0
    ):
        response_body = next(self.responses)
        return {"choices": [{"message": {"content": response_body}}]}, messages

    def count_tokens(self, text: Union[str, List[Dict[str, Any]]]) -> int:
        if isinstance(text, str):
            return len(text)
        elif isinstance(text, list):
            return sum(self.count_tokens(item["content"]) for item in text)
        else:
            raise ValueError(f"Unsupported text type: {type(text)}")

    def max_context_tokens(self) -> int:
        return -1



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple, Union


class BaseLLMClient(ABC):
    """Base class for all LLM clients that defines the standard interface."""

    @abstractmethod
    async def generate_completion(
        self,
        messages: List[Dict[str, str]],
        tools: Optional[List[Dict[str, Any]]] = None,
        temperature: float = 0.1,
        max_tokens: int = 8192,
        model: Optional[str] = None,
    ) -> Tuple[Dict[str, Any], List[Dict[str, Any]]]:
        """Generate a completion from the LLM.

        Args:
            messages: The messages to send to the LLM
            tools: Optional tools to provide to the LLM
            temperature: The temperature to use for generation
            max_tokens: The maximum number of tokens to generate
            model: Optional model override

        Returns:
            A tuple of (response, messages)
        """
        pass

    @abstractmethod
    async def generate_text(
        self,
        system_prompt: str,
        user_prompt: str,
        temperature: float = 0.1,
    ) -> str:
        """Generate text with a simple system and user prompt.

        Args:
            system_prompt: The system prompt
            user_prompt: The user prompt
            temperature: The temperature to use for generation

        Returns:
            The generated text
        """
        pass

    @abstractmethod
    def count_tokens(self, text_or_messages: Union[str, List[Dict[str, Any]]]) -> int:
        """Count tokens in a message list or string.

        Args:
            text_or_messages: Either a list of messages or a string

        Returns:
            The token count
        """
        pass

    @abstractmethod
    def max_context_tokens(self) -> int:
        """Get the maximum context size for the model.

        Returns:
            The maximum number of tokens the model can handle
            -1 means no limit
        """
        pass



import csv

from ai_migrate.manifest import Manifest


def main(*files: str, results_file: str = "results.csv"):
    writer = csv.writer(open(results_file, "w"))
    writer.writerow(["manifest_file", "time", "file", "result"])
    for manifest_file in files:
        print(f"Processing {manifest_file}")
        manifest = Manifest.model_validate_json(open(manifest_file).read())
        for file in manifest.files:
            writer.writerow(
                (
                    manifest_file,
                    manifest.time,
                    file.filename,
                    file.result,
                )
            )



import tempfile
import json
import subprocess
from pathlib import Path

import pytest

from ai_migrate.migrate import _run
from ai_migrate.fake_llm_client import FakeLLMClient
from ai_migrate.eval_generator import (
    generate_eval_from_migration,
    generate_eval_from_pr,
)


@pytest.fixture
def temp_project_dir():
    """Create a temporary project directory with examples and evals folders."""
    with tempfile.TemporaryDirectory() as temp_dir:
        project_dir = Path(temp_dir) / "project"
        examples_dir = project_dir / "examples"
        evals_dir = project_dir / "evals"

        examples_dir.mkdir(parents=True)
        evals_dir.mkdir(parents=True)

        old_file = examples_dir / "example.old.py"
        new_file = examples_dir / "example.new.py"

        with open(old_file, "w") as f:
            f.write('def old_function():\n    print("This is the old version")\n')

        with open(new_file, "w") as f:
            f.write('def new_function():\n    print("This is the new version")\n')

        system_prompt_file = project_dir / "system_prompt.md"
        with open(system_prompt_file, "w") as f:
            f.write("You are an expert at migrating code. Follow these patterns.")

        yield project_dir


@pytest.fixture
def temp_worktree():
    """Create a temporary directory to simulate a git worktree."""
    with tempfile.TemporaryDirectory() as temp_dir:
        worktree_dir = Path(temp_dir) / "worktree"
        worktree_dir.mkdir()

        test_file = worktree_dir / "example.py"
        with open(test_file, "w") as f:
            f.write(
                'def old_function():\n    print("This is the original version")\n    return "Old implementation"\n'
            )

        yield worktree_dir


@pytest.fixture
def fake_llm_client():
    """Create a FakeLLMClient with test responses."""
    responses_dir = Path(__file__).parent / "test" / "eval_test_data"
    return FakeLLMClient(responses_dir)


@pytest.mark.asyncio
async def test_eval_generation_with_flag(
    monkeypatch, temp_project_dir, temp_worktree, fake_llm_client
):
    """Test that evaluations are not created when the dont_create_evals flag is True."""

    monkeypatch.setattr("ai_migrate.migrate.DefaultClient", lambda: fake_llm_client)

    async def mock_subprocess_run(*args, **kwargs):
        return "mocked output"

    monkeypatch.setattr("ai_migrate.migrate.subprocess_run", mock_subprocess_run)

    class MockProcess:
        returncode = 0
        stdout = None

        async def communicate(self):
            return b"", b""

    async def mock_create_subprocess_exec(*args, **kwargs):
        return MockProcess()

    monkeypatch.setattr("asyncio.create_subprocess_exec", mock_create_subprocess_exec)

    target_file = temp_worktree / "example.py"
    with open(target_file, "w") as f:
        f.write('def original_function():\n    print("Original code")\n')

    verify_cmd = "echo success"

    await _run(
        target_files=[target_file],
        system_prompt=temp_project_dir / "system_prompt.md",
        examples_dir=temp_project_dir / "examples",
        verify_cmd=verify_cmd,
        pre_verify_cmd=None,
        worktree_root=temp_worktree,
        llm_fakes=None,
        dont_create_evals=True,
    )

    evals_dir = temp_project_dir / "evals"
    assert len(list(evals_dir.iterdir())) == 0


@pytest.mark.asyncio
async def test_eval_generation_without_flag(
    monkeypatch, temp_project_dir, temp_worktree, fake_llm_client
):
    """Test that evaluations are created when the dont_create_evals flag is False."""

    monkeypatch.setattr("ai_migrate.migrate.DefaultClient", lambda: fake_llm_client)

    async def mock_subprocess_run(*args, **kwargs):
        return "mocked output"

    monkeypatch.setattr("ai_migrate.migrate.subprocess_run", mock_subprocess_run)

    class MockProcess:
        returncode = 0
        stdout = None

        async def communicate(self):
            return b"", b""

    async def mock_create_subprocess_exec(*args, **kwargs):
        return MockProcess()

    monkeypatch.setattr("asyncio.create_subprocess_exec", mock_create_subprocess_exec)

    target_file = temp_worktree / "example.py"
    with open(target_file, "w") as f:
        f.write('def original_function():\n    print("Original code")\n')

    verify_cmd = "echo success"

    await _run(
        target_files=[target_file],
        system_prompt=temp_project_dir / "system_prompt.md",
        examples_dir=temp_project_dir / "examples",
        verify_cmd=verify_cmd,
        pre_verify_cmd=None,
        worktree_root=temp_worktree,
        llm_fakes=None,
        dont_create_evals=False,
    )

    evals_dir = temp_project_dir / "evals"
    eval_dirs = list(evals_dir.iterdir())
    assert len(eval_dirs) == 1

    eval_dir = eval_dirs[0]
    assert (eval_dir / "source").exists()
    assert (eval_dir / "manifest.json").exists()

    source_files = list((eval_dir / "source").iterdir())
    assert len(source_files) == 1
    assert source_files[0].name == "example.py"

    with open(source_files[0]) as f:
        pass

    with open(eval_dir / "manifest.json") as f:
        manifest = json.load(f)

        assert "files" in manifest
        assert len(manifest["files"]) > 0
        assert manifest["verify_cmd"] == verify_cmd


def test_generate_eval_from_migration():
    """Test the generate_eval_from_migration function directly."""
    with tempfile.TemporaryDirectory() as temp_dir:
        project_dir = Path(temp_dir)
        evals_dir = project_dir / "evals"
        evals_dir.mkdir()

        source_files = {"example.py": 'def old_function():\n    print("Original")\n'}
        transformed_files = {
            "example.py": 'def new_function():\n    print("Transformed")\n'
        }

        from ai_migrate.manifest import Manifest, FileEntry

        manifest = Manifest(
            files=[FileEntry(filename="example.py", result="pass")],
            verify_cmd="echo success",
        )

        eval_dir = generate_eval_from_migration(
            project_dir, source_files, transformed_files, manifest
        )

        assert eval_dir.exists()
        assert (eval_dir / "source").exists()
        assert (eval_dir / "manifest.json").exists()

        with open(eval_dir / "source" / "example.py") as f:
            content = f.read()
        assert content == source_files["example.py"]

        with open(eval_dir / "manifest.json") as f:
            saved_manifest = json.load(f)

        assert "files" in saved_manifest
        assert len(saved_manifest["files"]) == 1
        assert saved_manifest["files"][0]["filename"] == "example.py"
        assert saved_manifest["files"][0]["result"] == "pass"
        assert saved_manifest["verify_cmd"] == "echo success"


@pytest.mark.asyncio
async def test_generate_eval_from_pr_error_handling(monkeypatch):
    """Test error handling in generate_eval_from_pr function."""
    with tempfile.TemporaryDirectory() as temp_dir:
        project_dir = Path(temp_dir)
        evals_dir = project_dir / "evals"
        evals_dir.mkdir()

        class MockProcess:
            def __init__(self, cmd):
                self.cmd = cmd

                if cmd[0] == "git" and cmd[1] == "clone":
                    self.returncode = 1
                    self._stderr = b"Failed to clone repository"
                else:
                    self.returncode = 0
                    if cmd[0] == "gh" and cmd[1] == "pr" and cmd[2] == "view":
                        self._stdout = json.dumps(
                            {
                                "title": "Test PR",
                                "headRepository": {
                                    "name": "test-repo",
                                    "owner": {"login": "test-user"},
                                },
                                "files": [{"path": "test.py"}],
                                "baseRefOid": "abc123",
                            }
                        ).encode()
                    else:
                        self._stdout = b""
                    self._stderr = b""

            async def communicate(self):
                return getattr(self, "_stdout", b""), getattr(self, "_stderr", b"")

        async def mock_create_subprocess_exec(*args, **kwargs):
            return MockProcess(args)

        monkeypatch.setattr(
            "asyncio.create_subprocess_exec", mock_create_subprocess_exec
        )

        with pytest.raises(subprocess.CalledProcessError):
            await generate_eval_from_pr("test-pr-url", str(project_dir))

        eval_dirs = list(evals_dir.iterdir())
        assert len(eval_dirs) > 0
        eval_dir = eval_dirs[0]
        assert eval_dir.exists()
        assert (eval_dir / "source").exists()
        assert (eval_dir / "manifest.json").exists()



import tiktoken
from typing import Any, Dict, List, Optional, Tuple, Union

from openai import AsyncOpenAI

GPT_VERSION = "gpt-4o"


class OpenAIClient:
    """A client for interacting with a large language model."""

    def __init__(self):
        self._openai_client = AsyncOpenAI()

    async def generate_completion(
        self,
        messages: List[Dict[str, str]],
        tools: Optional[List[Dict[str, Any]]] = None,
        temperature: float = 0.1,
        max_tokens: int = 8192,
    ) -> Tuple[Dict[str, Any], List[Dict[str, Any]]]:
        """Generate a completion

        Args:
            messages: The messages to send
            tools: Optional tools to provide
            temperature: The temperature to use for generation
            max_tokens: The maximum number of tokens to generate

        Returns:
            A tuple of (response, messages)
        """
        response = await self._openai_client.chat.completions.create(
            model="gpt-4o",
            messages=messages,
            temperature=temperature,
            tools=tools,
            max_tokens=max_tokens,
        )
        response = response.model_dump()

        return response, messages

    async def generate_text(
        self,
        system_prompt: str,
        user_prompt: str,
        temperature: float = 0.1,
    ) -> str:
        """Generate text with a simple system and user prompt.

        Args:
            system_prompt: The system prompt
            user_prompt: The user prompt
            temperature: The temperature to use for generation

        Returns:
            The generated text
        """
        messages = [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": user_prompt},
        ]

        response, _ = await self.generate_completion(messages, temperature=temperature)
        return response["choices"][0]["message"]["content"]

    def count_tokens(self, text: Union[str, List[Dict[str, Any]]]) -> int:
        """Count the number of tokens in a string."""
        if isinstance(text, str):
            return len(tiktoken.encoding_for_model(GPT_VERSION).encode(text))
        elif isinstance(text, list):
            return sum(self.count_tokens(item["content"]) for item in text)
        else:
            raise ValueError(f"Unsupported text type: {type(text)}")

    def max_context_tokens(self) -> int:
        """Get the maximum context size for the model. -1 means no limit."""
        return -1
